# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""CIF file writer.

This module contains tools for writing `CIF <https://www.iucr.org/resources/cif>`_
files with diffraction data.
`It does not support reading CIF files.`

CIF Builder: High-level interface
---------------------------------
This module supports two interfaces for writing files.
The high-level interface uses a builder pattern for assembling files.
It is implemented by :class:`CIF`.

To demonstrate the interface, first make some mockup powder diffraction data:

  >>> import scipp as sc
  >>> tof = sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')
  >>> intensity = sc.array(
  ...     dims=['tof'],
  ...     values=[13.6, 26.0, 9.7],
  ...     variances=[0.7, 1.1, 0.5],
  ... )
  >>> da = sc.DataArray(intensity, coords={'tof': tof})

Assemble data and metadata using a :class:`CIF` builder:

  >>> from scippneutron.io import cif
  >>> cif_ = (
  ...  cif.CIF('my-data', comment="This is a demo of ScippNeutron's CIF builder.")
  ...  .with_beamline(beamline='fake', facility='made up')
  ...  .with_authors(cif.Author(name='Jane Doe', orcid='0000-0000-0000-0001'))
  ...  .with_reduced_powder_data(da)
  ... )

When all data has been added, write it to a file:

  >>> cif_.save('example.cif')

This results in a file containing

.. code-block:: text

    #\\#CIF_1.1
    # This is a demo of ScippNeutron's CIF builder.
    data_my-data

    loop_
    _audit_conform.dict_name
    _audit_conform.dict_version
    _audit_conform.dict_location
    pdCIF 2.5.0 https://github.com/COMCIFS/Powder_Dictionary/blob/7608b92165f58f968f054344e67662e01d4b401a/cif_pow.dic
    coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/fc3d75a298fd7c0c3cde43633f2a8616e826bfd5/cif_core.dic

    _audit.creation_date 2024-09-05T13:47:54+00:00
    _audit.creation_method 'Written by scippneutron v24.6.1

    _audit_contact_author.name 'Jane Doe'
    _audit_contact_author.id_orcid 0000-0000-0000-0001

    _diffrn_radiation.probe neutron
    _diffrn_source.beamline fake
    _diffrn_source.facility 'made up'

    loop_
    _pd_meas.time_of_flight
    _pd_proc.intensity_net
    _pd_proc.intensity_net_su
    1.2 13.6 0.8366600265340756
    1.4 26.0 1.0488088481701516
    2.3 9.7 0.7071067811865476

Chunks and loops: Low-level interface
-------------------------------------
The high-level CIF builder uses :class:`Block`, :class:`Chunk`, and :class:`Loop` to
encode data.
Those classes can also be used directly to gain more control over the file contents.

To demonstrate, make mockup powder diffraction data:

  >>> import scipp as sc
  >>> tof = sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')
  >>> intensity = sc.array(
  ...     dims=['tof'],
  ...     values=[13.6, 26.0, 9.7],
  ...     variances=[0.7, 1.1, 0.5],
  ... )

Wrap the data in a ``Loop`` to write them together as columns.

  >>> from scippneutron.io import cif
  >>> tof_loop = cif.Loop({
  ...     'pd_meas.time_of_flight': tof,
  ...     'pd_meas.intensity_total': sc.values(intensity),
  ...     'pd_meas.intensity_total_su': sc.stddevs(intensity),
  ... })

Write the data to file along with some metadata:

  >>> block = cif.Block('example', [
  ...     {
  ...         'diffrn_radiation.probe': 'neutron',
  ...         'diffrn_source.beamline': 'some-beamline',
  ...     },
  ...     tof_loop,
  ... ])
  >>> cif.save_cif('example.cif', block)

This results in a file containing

.. code-block::

  #\\#CIF_1.1
  data_example

  _diffrn_radiation.probe neutron
  _diffrn_source.beamline some-beamline

  loop_
  _pd_meas.time_of_flight
  _pd_meas.intensity_total
  _pd_meas.intensity_total_su
  1.2 13.6 0.8366600265340756
  1.4 26.0 1.0488088481701516
  2.3 9.7 0.7071067811865476
"""

from __future__ import annotations

import io
import warnings
from collections.abc import Iterable, Iterator, Mapping
from contextlib import contextmanager
from copy import copy
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

import scipp as sc


@dataclass(frozen=True)
class CIFSchema:
    name: str
    version: str
    location: str


CORE_SCHEMA = CIFSchema(
    name='coreCIF',
    version='3.3.0',
    location='https://github.com/COMCIFS/cif_core/blob/6f8502e81b623eb0fd779c79efaf191d49fa198c/cif_core.dic',
)
PD_SCHEMA = CIFSchema(
    name='pdCIF',
    version='2.5.0',
    location='https://github.com/COMCIFS/Powder_Dictionary/blob/970c2b2850a923796db5f4e9b7207d10ab5fd8e5/cif_pow.dic',
)


def save_cif(
    fname: str | Path | io.TextIOBase,
    content: Block | Iterable[Block] | CIF,
    *,
    comment: str = '',
) -> None:
    """Save data blocks to a CIF file.

    To use, first create :class:`scippneutron.io.cif.Block` objects to collect and
    structure data for the file, then use this function to write the file.

    Parameters
    ----------
    fname:
        Path or file handle for the output file.
    content:
        One or more CIF data blocks or a ``CIF`` object to write to the file.
    comment:
        Optional comment that can be written at the top of the file.
    """
    if isinstance(content, CIF):
        if comment:
            content = copy(content)
            content.comment = comment
        content.save(fname)
        return

    if isinstance(content, Block):
        content = (content,)
    with _open(fname) as f:
        _write_file_heading(f, comment=comment)
        _write_multi(f, content)


class CIF:
    """A builder for CIF files.

    This class implements a builder pattern for defining the contents of a CIF file.
    See the module documentation of :mod:`cif` for examples.
    """

    def __init__(self, name='', *, comment: str = '') -> None:
        self._block = Block(
            name=name
        )  # We mainly keep this around for managing the name.

        self._comment = ''
        self.comment = comment
        # Keep a separate list from self._block to assemble items
        # in a specific order when saving.
        self._content: list[Chunk | Loop] = []
        self._authors: list[Author] = []
        self._reducers: list[str] = []

        # Should be long enough to never run out of IDs.
        self._id_generator = (str(i) for i in range(1, 1_000_000_000))

    @property
    def name(self) -> str:
        return self._block.name

    @name.setter
    def name(self, name: str) -> None:
        self._block.name = name

    @property
    def comment(self) -> str:
        """Optional comment that can be written at the top of the file."""
        return self._comment

    @comment.setter
    def comment(self, comment: str) -> None:
        self._comment = _encode_non_ascii(comment)

    @property
    def schema(self) -> set[CIFSchema]:
        """CIF schemas used for the file."""
        return self._block.schema

    def save(self, fname: str | Path | io.TextIOBase) -> None:
        """
        Parameters
        ----------
        fname:
            Path or file handle for the output file.
        """
        block = self._block.copy()
        _add_audit(block, self._reducers)
        for item in (*self._assemble_authors(), *self._content):
            block.add(item)
        save_cif(fname, block, comment=self._comment)

    def copy(self) -> CIF:
        """Return a copy of this builder."""
        cif_ = CIF(name=self.name, comment=self.comment)
        cif_._block = self._block.copy()
        cif_._content.extend(self._content)
        cif_._authors.extend(self._authors)
        cif_._reducers.extend(self._reducers)
        return cif_

    # TODO use Beamline model when available
    #  see https://github.com/scipp/scippneutron/issues/473
    def with_beamline(
        self,
        *,
        beamline: str,
        facility: str | None = None,
        device: str | None = None,
        comment: str = '',
    ) -> CIF:
        """Add beamline information."""
        if device is None:
            if (facility or '').lower() in _KNOWN_SPALLATION_SOURCES:
                device = 'spallation'
        fields = {
            'diffrn_radiation.probe': 'neutron',
            'diffrn_source.beamline': beamline,
            'diffrn_source.facility': facility,
            'diffrn_source.device': device,
        }

        cif_ = self.copy()
        cif_._content.append(
            Chunk(
                {k: v for k, v in fields.items() if v is not None},
                comment=comment,
                schema=CORE_SCHEMA,
            )
        )
        return cif_

    def with_reduced_powder_data(self, data: sc.DataArray, *, comment: str = '') -> CIF:
        """Add a loop with reduced powder data.

        The input must be 1-dimensional with a dimension name in
        ``('tof', 'dspacing')``.
        The data array may also have a name in
        ``('intensity_net', 'intensity_norm', 'intensity_total')``.
        If the name is not set, it defaults to ``'intensity_norm'``,
        which is appropriate for typical outputs from data reduction.
        See https://github.com/COMCIFS/Powder_Dictionary/blob/master/cif_pow.dic

        The data gets written as intensity along a single coord whose
        name matches the dimension name.
        Standard uncertainties are also written if present.

        The unit of the coordinate must match the requirement of pdCIF.

        Parameters
        ----------
        data:
            1-dimensional data array with a recognized dimension name
        comment:
            Optional comment that can be written above the data in the file.

        Returns
        -------
        :
            A builder with added reduced data.

        Examples
        --------
        Make mockup powder diffraction data:

          >>> import scipp as sc
          >>> tof = sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')
          >>> intensity = sc.array(
          ...     dims=['tof'],
          ...     values=[13.6, 26.0, 9.7],
          ...     variances=[0.7, 1.1, 0.5],
          ... )

        Add to a CIF builder:

          >>> from scippneutron.io import cif
          >>> cif_ = cif.CIF('reduced-data')
          >>> da = sc.DataArray(intensity, coords={'tof': tof})
          >>> cif_ = cif_.with_reduced_powder_data(da)
        """
        cif_ = self.copy()
        cif_._content.append(_make_reduced_powder_loop(data, comment=comment))
        return cif_

    def with_powder_calibration(self, cal: sc.DataArray, *, comment: str = '') -> CIF:
        r"""Add a powder calibration table.

        The calibration data encode the following transformation from
        d-spacing to time-of-flight:

        .. math::

            t = \sum_{i=0}^N\, c_i d^{p_i}

        where :math:`c_i` is the i-th element of ``cal`` and :math:`p^{p_i}`
        is the i-th element of ``cal.coords['power']``.

        Parameters
        ----------
        cal:
            The data are the calibration coefficients (possibly with variances).
            Must have a coordinate called ``'power'`` defining :math:`p` in the
            equation above.
        comment:
            Optional comment that can be written above the data in the file.

        Returns
        -------
        :
            A builder with added calibration data.

        Examples
        --------
        Add a mockup calibration table:

          >>> import scipp as sc
          >>> from scippneutron.io import cif
          >>> cal = sc.DataArray(
          ...     sc.array(dims=['cal'], values=[3.4, 0.2]),
          ...     coords={'power': sc.array(dims=['cal'], values=[0, 1])},
          ... )
          >>> cif_ = cif.CIF('powder-calibration')
          >>> cif_ = cif_.with_powder_calibration(cal)
        """
        cif_ = self.copy()
        cif_._content.append(_make_powder_calibration_loop(cal, comment=comment))
        return cif_

    def with_authors(self, *authors: Author) -> CIF:
        """Add one or more authors.

        Parameters
        ----------
        authors:
            Authors to add to this CIF object.

        Returns
        -------
        :
            A builder with added calibration data.
        """
        cif_ = self.copy()
        cif_._authors.extend(authors)
        return cif_

    def with_reducers(self, *reducers: str) -> CIF:
        """Add one or more programs that were used to reduce the software.

        Parameters
        ----------
        reducers:
            Pieces of software that were used to reduce the data.
            Each string is one piece of software in freeform notation.
            It is recommended to include the program name and version.

        Returns
        -------
        :
            A builder with added calibration data.
        """
        cif_ = self.copy()
        cif_._reducers.extend(reducers)
        return cif_

    def _assemble_authors(self) -> list[Chunk | Loop]:
        contact = [author for author in self._authors if author.corresponding]
        regular = [author for author in self._authors if not author.corresponding]

        results = []
        roles = {}
        for authors, category in zip(
            (contact, regular), ('audit_contact_author', 'audit_author'), strict=True
        ):
            if not authors:
                continue
            data, rols = _serialize_authors(authors, category, self._id_generator)
            results.append(data)
            roles.update(rols)
        if roles:
            results.append(_serialize_roles(roles))

        return results


# TODO replace with common metadata `Person` once that exists.
#  See https://github.com/scipp/scippneutron/issues/473
@dataclass(kw_only=True)
class Author:
    name: str
    address: str | None = None
    email: str | None = None
    orcid: str | None = None
    corresponding: bool = True
    role: str | None = None


class _CIFBase:
    def __init__(self, *, comment: str = '', schema: CIFSchema) -> None:
        self._comment = ''
        self.comment = comment  # use str-encoding logic
        self._schema = _preprocess_schema(schema)

    @property
    def comment(self) -> str:
        """Optional comment that can be written above the object in the file."""
        return self._comment

    @comment.setter
    def comment(self, comment: str) -> None:
        self._comment = _encode_non_ascii(comment)

    @property
    def schema(self) -> set[CIFSchema]:
        """CIF schemas used for the object."""
        return self._schema


class Chunk(_CIFBase):
    """A group of CIF key-value pairs.

    Chunks contain one or more key-value pairs where values are scalars,
    i.e., not arrays.
    Chunks are represented in files as a group of pairs separated from
    other chunks and loops by empty lines.

    Note that CIF has no concept of chunks; they are only used for organizing
    data in ScippNeutron.
    """

    def __init__(
        self,
        pairs: Mapping[str, Any] | Iterable[tuple[str, Any]] | None,
        /,
        *,
        comment: str = '',
        schema: CIFSchema | Iterable[CIFSchema] | None = None,
    ) -> None:
        """Create a new CIF chunk.

        Parameters
        ----------
        pairs:
            Defines a mapping from keys (a.k.a. tags) to values.
        comment:
            Optional comment that can be written above the chunk in the file.
        schema:
            CIF schemas used for the chunk.
            Content is not checked against the schema, but the schema is written
            to the file.
        """
        super().__init__(comment=comment, schema=schema)
        self._pairs = dict(pairs) if pairs is not None else {}

    def __setitem__(self, key: str, value: Any) -> None:
        """Add a key-value pair to the chunk."""
        self._pairs[key] = value

    def write(self, f: io.TextIOBase) -> None:
        """Write this chunk to a file.

        Used mainly internally, use :func:`scippneutron.io.cif.save_cif` instead.

        Parameters
        ----------
        f:
            File handle.
        """
        _write_comment(f, self.comment)
        for key, val in self._pairs.items():
            v = _format_value(val)
            if v.startswith(';'):
                f.write(f'_{key}\n{v}\n')
            else:
                f.write(f'_{key} {v}\n')


class Loop(_CIFBase):
    """A CIF loop.

    Contains a mapping from strings to Scipp variables.
    The strings are arbitrary and ``Loop`` can merge items from different categories
    into a single loop.
    All variables must have the same length.
    """

    def __init__(
        self,
        columns: Mapping[str, sc.Variable] | Iterable[tuple[str, sc.Variable]] | None,
        *,
        comment: str = '',
        schema: CIFSchema | Iterable[CIFSchema] | None = None,
    ) -> None:
        """Create a new CIF loop.

        Parameters
        ----------
        columns:
            Defines a mapping from column names (including their category)
            to column values as Scipp variables.
        comment:
            Optional comment that can be written above the loop in the file.
        schema:
            CIF schemas used for the loop.
            Content is not checked against the schema, but the schema is written
            to the file.
        """
        super().__init__(comment=comment, schema=schema)
        self._columns = {}
        for key, column in columns.items():
            self[key] = column

    def __setitem__(self, name: str, value: sc.Variable) -> None:
        """Add a column to the loop.

        Parameters
        ---------
        name:
            Column name.
        value:
            Values of the column.
        """
        if value.ndim != 1:
            raise sc.DimensionError(
                "CIF loops can only contain 1d variables, got " f"{value.ndim} dims"
            )
        if self._columns:
            existing = next(iter(self._columns.values())).sizes
            if existing != value.sizes:
                raise sc.DimensionError(
                    f"Inconsistent dims in CIF loop: {value.sizes} "
                    f"loop dims: {existing}"
                )

        self._columns[name] = value

    def write(self, f: io.TextIOBase) -> None:
        """Write this loop to a file.

        Used mainly internally, use :func:`scippneutron.io.cif.save_cif` instead.

        Parameters
        ----------
        f:
            File handle.
        """
        _write_comment(f, self.comment)
        f.write('loop_\n')
        for key in self._columns:
            f.write(f'_{key}\n')
        formatted_values = [
            tuple(map(_format_value, row))
            for row in zip(*self._columns.values(), strict=True)
        ]
        # If any value is a multi-line string, lay out elements as a flat vertical
        # list, otherwise use a 2d table.
        sep = (
            '\n'
            if any(';' in item for row in formatted_values for item in row)
            else ' '
        )
        for row in formatted_values:
            f.write(sep.join(row))
            f.write('\n')


class Block(_CIFBase):
    """A CIF data block.

    A block contains an ordered sequence of loops
    and chunks (groups of key-value-pairs).
    The contents are written to file in the order specified in the block.
    """

    def __init__(
        self,
        name: str,
        content: Iterable[Mapping[str, Any] | Loop | Chunk] | None = None,
        *,
        comment: str = '',
        schema: CIFSchema | Iterable[CIFSchema] | None = None,
    ) -> None:
        """Create a new CIF data block.

        Parameters
        ----------
        name:
            Name of the block.
            Can contain any non-whitespace characters.
            Can be at most 75 characters long.
        content:
            Initial loops and chunks.
            ``dicts`` are converted to :class:`scippneutron.io.cif.Chunk`.
        comment:
            Optional comment that can be written above the block in the file.
        schema:
            CIF schemas used for the block.
            Content is not checked against the schema, but the schema is written
            to the file.
        """
        super().__init__(comment=comment, schema=schema)
        self._name = ''
        self.name = name
        self._content = _convert_input_content(content) if content is not None else []

    @property
    def name(self) -> str:
        """Name of the block."""
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = _encode_non_ascii(name)
        if ' ' in self._name or '\t' in self._name or '\n' in self._name:
            raise ValueError(
                "Block name must not contain spaces or line breaks, "
                f"got: '{self._name}'"
            )
        if len(self._name) > 75:
            warnings.warn(
                "cif.Block name should not be longer than 75 characters, got "
                f"{len(self._name)} characters ('{self._name}')",
                UserWarning,
                stacklevel=2,
            )

    @property
    def schema(self) -> set[CIFSchema]:
        """CIF schemas used for the block."""
        merged = set(super().schema)
        for item in self._content:
            merged.update(item.schema)
        return merged

    def copy(self) -> Block:
        """Return a shallow copy of the block."""
        return Block(
            self.name, list(self._content), comment=self.comment, schema=self.schema
        )

    def add(
        self,
        content: Mapping[str, Any] | Iterable[tuple[str, Any]] | Chunk | Loop,
        /,
        comment: str = '',
    ) -> None:
        """Add a chunk or loop to the end of the block.

        Parameters
        ----------
        content:
            A loop, chunk, or mapping to add.
            Mappings get converted to chunks.
        comment:
            Optional comment that can be written above the chunk or loop in the file.
        """
        if not isinstance(content, Chunk | Loop):
            content = Chunk(content, comment=comment)
        self._content.append(content)

    def write(self, f: io.TextIOBase) -> None:
        """Write this block to a file.

        Used mainly internally, use :func:`scippneutron.io.cif.save_cif` instead.

        Parameters
        ----------
        f:
            File handle.
        """
        schema_loop = _make_schema_loop(self.schema)

        _write_comment(f, self.comment)
        f.write(f'data_{self.name}\n\n')
        if schema_loop is not None:
            schema_loop.write(f)
            f.write('\n')
        _write_multi(f, self._content)


_KNOWN_SPALLATION_SOURCES = {
    'csns',
    'ess',
    'isis',
    'j-parc',
    'lansce' 'sinq',
    'sns',
}


def _convert_input_content(
    content: Iterable[Mapping[str, Any] | Loop | Chunk],
) -> list[Loop | Chunk]:
    return [item if isinstance(item, Loop | Chunk) else Chunk(item) for item in content]


@contextmanager
def _open(fname: str | Path | io.TextIOBase):
    if isinstance(fname, io.TextIOBase):
        yield fname
    else:
        with open(fname, 'w') as f:
            yield f


def _preprocess_schema(
    schema: CIFSchema | Iterable[CIFSchema] | None,
) -> set[CIFSchema]:
    if schema is None:
        return set()
    if isinstance(schema, CIFSchema):
        res = {schema}
    else:
        res = set(schema)
    res.add(CORE_SCHEMA)  # needed to encode schema itself
    return res


def _make_schema_loop(schema: set[CIFSchema]) -> Loop | None:
    if not schema:
        return None
    columns = {
        'audit_conform.dict_name': [],
        'audit_conform.dict_version': [],
        'audit_conform.dict_location': [],
    }
    for s in schema:
        columns['audit_conform.dict_name'].append(s.name)
        columns['audit_conform.dict_version'].append(s.version)
        columns['audit_conform.dict_location'].append(s.location)
    return Loop(
        {key: sc.array(dims=['schema'], values=val) for key, val in columns.items()}
    )


def _quotes_for_string_value(value: str) -> str | None:
    if '\n' in value:
        return ';'
    if "'" in value:
        if '"' in value:
            return ';'
        return '"'
    if '"' in value:
        return "'"
    if ' ' in value:
        return "'"
    if not value:
        return "'"  # so that empty strings are shown as ''
    return None


def _encode_non_ascii(s: str) -> str:
    return s.encode('ascii', 'backslashreplace').decode('ascii')


def _format_value(value: Any) -> str:
    if isinstance(value, sc.Variable):
        if value.variance is not None:
            without_unit = sc.scalar(value.value, variance=value.variance)
            s = f'{without_unit:c}'
        else:
            s = str(value.value)
    elif isinstance(value, datetime):
        s = value.isoformat()
    else:
        s = str(value)

    s = _encode_non_ascii(s)

    if (quotes := _quotes_for_string_value(s)) == ';':
        return f'; {s}\n;'
    elif quotes is not None:
        return quotes + s + quotes
    return s


def _write_comment(f: io.TextIOBase, comment: str) -> None:
    if comment:
        f.write('# ')
        f.write('\n# '.join(comment.splitlines()))
        f.write('\n')


def _write_multi(f: io.TextIOBase, to_write: Iterable[Any]) -> None:
    first = True
    for item in to_write:
        if not first:
            f.write('\n')
        first = False
        item.write(f)


def _write_file_heading(f: io.TextIOBase, comment: str) -> None:
    f.write('#\\#CIF_1.1\n')
    _write_comment(f, comment)


def _reduced_powder_coord(data) -> tuple[str, sc.Variable]:
    if data.ndim != 1:
        raise sc.DimensionError(f'Can only save 1d powder data, got {data.ndim} dims.')
    known_coords = {
        'tof': ('pd_meas.time_of_flight', 'us'),
        'dspacing': ('pd_proc.d_spacing', 'Å'),
    }
    try:
        name, unit = known_coords[data.dim]
    except KeyError:
        raise sc.CoordError(
            f'Unrecognized dim: {data.dim}. Must be one of {list(known_coords)}'
        ) from None

    coord = data.coords[data.dim]
    if coord.unit != unit:
        raise sc.UnitError(
            f'Incorrect unit for powder coordinate {name}: {coord.unit} '
            f'expected {unit}'
        )
    return name, coord


def _normalize_reduced_powder_name(name: str) -> str:
    if name not in ('intensity_net', 'intensity_norm', 'intensity_total'):
        raise ValueError(f'Unrecognized name for reduced powder data: {name}')
    return f'pd_proc.{name}'


def _make_reduced_powder_loop(data: sc.DataArray, comment: str) -> Loop:
    coord_name, coord = _reduced_powder_coord(data)
    data_name = _normalize_reduced_powder_name(data.name or 'intensity_norm')

    res = Loop(
        {
            'pd_data.point_id': sc.arange(data.dim, len(data), unit=None),
            coord_name: sc.values(coord),
        },
        comment=comment,
        schema=PD_SCHEMA,
    )
    if coord.variances is not None:
        res[coord_name + '_su'] = sc.stddevs(coord)
    res[data_name] = sc.values(data.data)
    if data.variances is not None:
        res[data_name + '_su'] = sc.stddevs(data.data)

    if data.unit != 'one':
        pre = res.comment + '\n' if res.comment else ''
        res.comment = f'{pre}Unit of intensity: [{data.unit}]'

    return res


def _make_powder_calibration_loop(data: sc.DataArray, comment: str) -> Loop:
    # All names are valid python identifiers
    id_by_power = {0: 'ZERO', 1: 'DIFC', 2: 'DIFA', -1: 'DIFB'}
    ids = sc.array(
        dims=[data.dim],
        values=[
            id_by_power.get(power, f'c{power}'.replace('-', '_').replace('.', '_'))
            for power in data.coords['power'].values
        ],
    )
    res = Loop(
        {
            'pd_calib_d_to_tof.id': ids,
            'pd_calib_d_to_tof.power': data.coords['power'],
            'pd_calib_d_to_tof.coeff': sc.values(data.data),
        },
        comment=comment,
        schema=PD_SCHEMA,
    )
    if data.variances is not None:
        res['pd_calib_d_to_tof.coeff_su'] = sc.stddevs(data.data)
    return res


def _add_audit(block: Block, reducers: list[str]) -> None:
    from .. import __version__

    audit_chunk = Chunk(
        {
            'audit.creation_date': datetime.now(timezone.utc).replace(microsecond=0),
            'audit.creation_method': f'Written by scippneutron v{__version__}',
        },
        schema=CORE_SCHEMA,
    )
    block.add(audit_chunk)

    if len(reducers) == 1:
        audit_chunk['computing.diffrn_reduction'] = reducers[0]
    elif len(reducers) > 1:
        block.add(
            Loop(
                {'computing.diffrn_reduction': sc.array(dims=['r'], values=reducers)},
                schema=CORE_SCHEMA,
            )
        )


def _serialize_authors(
    authors: list[Author],
    category: str,
    id_generator: Iterator[str],
) -> tuple[Chunk | Loop, dict[str, str]]:
    fields = {
        f'{category}.{key}': f
        for key in ('name', 'email', 'address', 'orcid')
        if any(f := [getattr(a, key) or '' for a in authors])
    }
    # Map between our name (Author.orcid) and CIF's (id_orcid)
    if orcid_id := fields.pop(f'{category}.orcid', None):
        fields[f'{category}.id_orcid'] = orcid_id

    roles = {next(id_generator): a.role for a in authors}
    if any(roles.values()):
        fields[f'{category}.id'] = list(roles.keys())
    roles = {key: val for key, val in roles.items() if val}

    if len(authors) == 1:
        return Chunk(
            {key: val[0] for key, val in fields.items()},
            schema=CORE_SCHEMA,
        ), roles
    return Loop(
        {key: sc.array(dims=['author'], values=val) for key, val in fields.items()},
        schema=CORE_SCHEMA,
    ), roles


def _serialize_roles(roles: dict[str, str]) -> Loop:
    return Loop(
        {
            'audit_author_role.id': sc.array(dims=['role'], values=list(roles)),
            'audit_author_role.role': sc.array(
                dims=['role'], values=list(roles.values())
            ),
        },
        schema=CORE_SCHEMA,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""
Internal utilities; do not use outside scippneutron!
"""

from collections.abc import MutableMapping

import scipp as sc
from scipp.typing import VariableLike


def elem_unit(var: VariableLike) -> sc.Unit:
    return var.bins.unit if var.bins is not None else var.unit


def elem_dtype(var: VariableLike) -> sc.DType:
    return var.bins.constituents['data'].dtype if var.bins is not None else var.dtype


def float_dtype(var: VariableLike) -> sc.DType:
    dtype = elem_dtype(var)
    if dtype == sc.DType.float32:
        return sc.DType.float32
    return sc.DType.float64


def as_float_type(var: VariableLike, ref: VariableLike) -> VariableLike:
    return var.astype(float_dtype(ref), copy=False)


def get_attrs(da: sc.DataArray) -> MutableMapping[str, sc.Variable]:
    try:
        # During deprecation phase
        return da.deprecated_attrs
    except AttributeError:
        try:
            # Before deprecation phase
            return da.attrs
        except AttributeError:
            # After deprecation phase / removal of attrs
            return da.coords


def get_meta(da: sc.DataArray) -> MutableMapping[str, sc.Variable]:
    try:
        # During deprecation phase
        return da.deprecated_meta
    except AttributeError:
        try:
            # Before deprecation phase
            return da.meta
        except AttributeError:
            # After deprecation phase / removal of attrs
            return da.coords



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Iterable

import scipp as sc

from ._fit_peaks import FitResult


def remove_peaks(data: sc.DataArray, fit_results: Iterable[FitResult]) -> sc.DataArray:
    """Remove peaks from data by subtracting fitted models.

    Peaks are removed by subtracting fitted peak models from the data.
    The background models are ignored in this operation.

    Parameters
    ----------
    data:
        1d data with peaks.
        Must have a dimension-coordinate.
    fit_results:
        Results of fitting the peaks that should be removed.
        Any unsuccessful fits are ignored.

    Returns
    -------
    :
        ``data`` with peaks removed.

    See Also
    --------
    fit_peaks:
        Fit the peaks in ``data`` and construct the required input to ``fit_results``.
    """
    if data.variances is not None:
        raise sc.VariancesError(
            'Cannot remove peaks from data with variances as that would introduce '
            'correlations between data points.'
        )

    data = data.copy(deep=False)
    data.data = data.data.copy()  # need a deep copy for the in-place subtraction below
    for result in fit_results:
        if not result.success:
            continue
        in_window = data[data.dim, result.window[0] : result.window[1]]
        in_window -= result.eval_peak(in_window.coords[data.dim])
    return data



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# flake8: noqa: E501

"""SVG rendering for choppers."""

import dataclasses
import math
from string import Template

import scipp as sc

from .disk_chopper import DiskChopper

_CHOPPER_TEMPLATE = Template(
    """<svg
    version="1.1"
    width="${image_size}"
    height="${image_size}"
    viewBox="0 0 ${image_size} ${image_size}"
    xmlns="http://www.w3.org/2000/svg">
${elements}
</svg>
"""
)
_DISK_TEMPLATE = Template(
    '<path d="${path}" fill="#e0e0e0" ' 'stroke="#a0a0a0" stroke-width="2"/>'
)
_EDGE_MARK_TEMPLATE = Template(
    '<path d="${path}" stroke="#707070" stroke-width="2" stroke-dasharray="5,5"/>'
)
_EDGE_LABEL_TEMPLATE = Template(
    '<text x="${x}" y="${y}" text-anchor="${anchor}" font-family="sans-serif" '
    'dominant-baseline="ideographic" fill="#707070">'
    "${label}</text>"
)
_TDC_MARK_TEMPLATE = Template(
    """<path d="${path}" stroke="#0065ac" stroke-width="2" stroke-dasharray="5,5"/>
<text x="${text_x}" y="${text_y}" text-anchor="middle" dominant-baseline="middle" font-family="sans-serif" fill="#0065ac">TDC</text>"""
)
_BEAM_POSITION_TEMPLATE = Template(
    """<path d="${path}" stroke="#c83737" stroke-width="2" stroke-dasharray="5,5"/>
<text x="${text_x}" y="${text_y}" text-anchor="middle" dominant-baseline="middle" font-family="sans-serif" fill="#c83737">beam position</text>"""
)


def _rotation_arrow(*, image_size: int, clockwise: bool) -> str:
    # The arrow was drawn with Inkscape at an image size of 400.
    s = image_size / 400
    flip, shift = (-1, image_size) if clockwise else (1, 0)
    return f"""<g id="rotation-arrow" transform="matrix({flip*s},0,0,{s},{shift},20)">
<path fill="#606060"
 d="m 190.46289,4.140625 c -19.2285,0.9736749 -38.20846,4.7790001 -56.32617,11.292969 l 0.89453,2.490234 C 152.90409,11.497902 171.62709,7.7437179 190.5957,6.7832031 Z"/>
<path fill="#606060" stroke-linecap="round"
 d="m 138.29883,5.5019531 a 1.322835,1.322835 0 0 0 -0.75,0.6816406 l -5.34766,11.3515623 11.35156,5.347656 A 1.322835,1.322835 0 0 0 145.3125,22.25 1.322835,1.322835 0 0 0 144.67969,20.488281 l -8.95703,-4.21875 4.21875,-8.9589841 a 1.322835,1.322835 0 0 0 -0.63282,-1.7597657 1.322835,1.322835 0 0 0 -1.00976,-0.048828 z"/>
</g>"""


def _combine_slits(chopper: DiskChopper) -> sc.DataArray:
    edges = sc.concat(
        [chopper.slit_begin.flatten(to="slit"), chopper.slit_end.flatten(to="slit")],
        dim="edge",
    )
    height = chopper.slit_height.rename_dims({chopper.slit_height.dim: "slit"})
    slits = sc.DataArray(
        sc.arange("slit", edges.sizes["slit"], unit=None),
        coords={
            "edge": edges,
            "height": height,
        },
    )
    return sc.sort(slits, key=slits.coords["edge"]["edge", 0])


def _preprocess(chopper: DiskChopper) -> DiskChopper:
    radius = sc.scalar(1.0, unit="m") if chopper.radius is None else chopper.radius
    slit_height = radius / 2 if chopper.slit_height is None else chopper.slit_height
    return dataclasses.replace(chopper, radius=radius, slit_height=slit_height)


def draw_disk_chopper(chopper: DiskChopper, *, image_size: int) -> str:
    """Generate an SVG image for a chopper."""

    # This functions and its sub-functions mimic an SVG path in that they maintain
    # a state that gets updated with every segment that gets added to the path.
    # This way, they trace out the chopper disk and its slits.
    #
    # There are two coordinate systems.
    # - Object coords: The chopper's own coordinate system in which its radius and
    #                  slit height are defined.
    #                  The origin is in the center of the chopper.
    # - SVG coords: The coords used for drawing in a box of size
    #               ``image_size x image_size``.
    #               The origin is in the top-left.
    #
    # ``scale`` maps from object to svg lengths.
    # In addition, ``radius_scale`` is applied to the disk radius to make room for
    # additional elements around the disk.
    # ``to_svg_coord`` maps a single x or y object coord to an SVG coord.
    # ``polar_to_svg_coords`` maps polar coordinates to cartesian SVG coordinates.
    # It applies a pi/2 counterclockwise rotation in order to position angle=0
    # at the top as chopper angles are defined in terms of the top dead center sensor.
    chopper = _preprocess(chopper)

    scale = image_size / 2 / chopper.radius * 0.99
    radius_scale = 0.75

    def to_svg_coord(x: sc.Variable) -> float:
        return float(x * scale) + image_size // 2

    def to_svg_length(x: sc.Variable) -> float:
        return float(x * scale)

    def polar_to_svg_coords(*, r: sc.Variable, a: sc.Variable) -> tuple[float, float]:
        x = to_svg_coord(r * -sc.sin(a))
        y = image_size - to_svg_coord(r * sc.cos(a))
        return x, y

    # The state of the disk tracer.
    angle = sc.scalar(0.0, unit="rad")
    outer = True  # Currently tracing the outer circle or slit?
    radius = chopper.radius * radius_scale

    def move_to(*, r: sc.Variable, a: sc.Variable) -> str:
        nonlocal radius, angle
        a = a.to(unit="rad")
        radius = r
        angle = a
        x, y = polar_to_svg_coords(r=r, a=a)
        return f"M{x:.3f} {y:.3f}"

    def trace_arc(a: sc.Variable) -> str:
        nonlocal angle
        a = a.to(unit="rad")
        x, y = polar_to_svg_coords(r=radius, a=a)
        r = to_svg_length(radius)
        pi = sc.scalar(math.pi, unit="rad")
        large = 1 if (a - angle) > pi else 0
        angle = a
        return f"A{r:.3f} {r:.3f} 0 {large} 0 {x:.3f} {y:.3f}"

    def trace_edge(h: sc.Variable) -> str:
        nonlocal outer, radius
        if outer:
            radius = (chopper.radius - h) * radius_scale
            outer = False
        else:
            radius = chopper.radius * radius_scale
            outer = True
        x, y = polar_to_svg_coords(r=radius, a=angle)
        return f"L{x:.3f} {y:.3f}"

    def edge_mark(is_begin: bool, idx: int) -> tuple[str, dict[str, str]]:
        x, y = polar_to_svg_coords(
            r=(chopper.radius - chopper.slit_height[idx]) * radius_scale, a=angle
        )
        line = f"M{image_size // 2} {image_size // 2} L{x:.3f} {y:.3f}"

        x, y = polar_to_svg_coords(r=chopper.radius * radius_scale, a=angle)
        anchor = "start" if x > image_size // 2 else "end"
        text = {
            "x": x,
            "y": y,
            "anchor": anchor,
            "label": ("begin" if is_begin else "end") + str(idx),
        }
        return line, text

    def tdc_marker() -> str:
        x, y = polar_to_svg_coords(
            r=chopper.radius * radius_scale, a=sc.scalar(0.0, unit="rad")
        )
        return _TDC_MARK_TEMPLATE.substitute(
            path=f"M{image_size // 2} {image_size // 2} L{x} {y}",
            text_x=x,
            text_y=y - 11,
        )

    def beam_pos(beam_position: sc.Variable) -> str:
        x, y = polar_to_svg_coords(r=chopper.radius * radius_scale, a=beam_position)
        return _BEAM_POSITION_TEMPLATE.substitute(
            path=f"M{image_size // 2} {image_size // 2} L{x} {y}",
            text_x=x,
            text_y=y,
        )

    def trace_disk() -> tuple[list[str], list[tuple[str, dict[str, str]]]]:
        slits = _combine_slits(chopper)
        if len(slits) == 0:
            return [], []

        start_angle = None
        path = []
        marks = []
        for slit in slits:
            begin, end = slit.coords["edge"]

            if start_angle is None:
                start_angle = begin.to(unit="rad")
                path.append(move_to(r=chopper.radius * radius_scale, a=start_angle))
            else:
                path.append(trace_arc(begin))
            path.append(trace_edge(slit.coords["height"]))
            marks.append(edge_mark(is_begin=True, idx=slit.value))
            path.append(trace_arc(end))
            path.append(trace_edge(slit.coords["height"]))
            marks.append(edge_mark(is_begin=False, idx=slit.value))

        if start_angle < angle:
            path.append(trace_arc(sc.scalar(2 * math.pi, unit="rad") + start_angle))
        else:
            path.append(trace_arc(start_angle))

        return path, marks

    disk_path, edge_marks = trace_disk()

    elements = [
        _DISK_TEMPLATE.substitute(path=" ".join(disk_path)),
        *(
            _EDGE_MARK_TEMPLATE.substitute(path=path)
            + "\n"
            + _EDGE_LABEL_TEMPLATE.substitute(**label)
            for path, label in edge_marks
        ),
        tdc_marker(),
        _rotation_arrow(image_size=image_size, clockwise=chopper.is_clockwise),
    ]
    if (beam_position := chopper.beam_position) is not None:
        elements.append(beam_pos(beam_position))
    elements.append(f'<circle cx="{image_size/2}" cy="{image_size/2}" r="5"/>')
    return _CHOPPER_TEMPLATE.substitute(
        image_size=image_size, elements="\n".join(elements)
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import json

import numpy as np
import pytest
import scipp as sc
from matplotlib.colors import to_hex

from scippneutron import MaskingTool


def make_data(npoints=50000, scale=10.0, seed=1):
    rng = np.random.default_rng(seed)
    position = scale * rng.standard_normal(size=[npoints, 2])
    values = np.linalg.norm(position, axis=1)
    return sc.DataArray(
        data=sc.array(dims=['row'], values=values, unit='K'),
        coords={
            'x': sc.array(dims=['row'], unit='m', values=position[:, 0]),
            'y': sc.array(dims=['row'], unit='m', values=position[:, 1]),
        },
    ).hist(y=300, x=300)


@pytest.mark.usefixtures('_use_ipympl')
def test_mask_2d_with_rectangles():
    da = make_data()
    masking_tool = MaskingTool(da)
    r, _, _ = masking_tool.controls

    r.value = True
    r._tool.click(-10.0, -10.0)
    r._tool.click(5.0, 6.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 1
    mask = next(iter(masks.values()))
    assert mask["kind"] == "rectangle"
    bounds = mask["bounds"]
    assert len(bounds) == 2  # 2 dimensions
    assert bounds['x']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 5.0, "unit": "m"}
    assert bounds['y']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 6.0, "unit": "m"}

    r._tool.click(10.0, -20.0)
    r._tool.click(15.0, 30.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 2
    mask = list(masks.values())[1]
    assert mask["kind"] == "rectangle"
    bounds = mask["bounds"]
    assert len(bounds) == 2  # 2 dimensions
    assert bounds['x']["min"] == {"value": 10.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 15.0, "unit": "m"}
    assert bounds['y']["min"] == {"value": -20.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 30.0, "unit": "m"}


@pytest.mark.usefixtures('_use_ipympl')
def test_mask_2d_with_vspans():
    da = make_data()
    masking_tool = MaskingTool(da)
    _, v, _ = masking_tool.controls

    v.value = True
    v._tool.click(-20.0, 0.0)
    v._tool.click(1.0, 0.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 1
    mask = next(iter(masks.values()))
    assert mask["kind"] == "vspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['x']["min"] == {"value": -20.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 1.0, "unit": "m"}

    v._tool.click(21.0, 0.0)
    v._tool.click(27.0, 0.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 2
    mask = list(masks.values())[1]
    assert mask["kind"] == "vspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['x']["min"] == {"value": 21.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 27.0, "unit": "m"}


@pytest.mark.usefixtures('_use_ipympl')
def test_mask_2d_with_hspans():
    da = make_data()
    masking_tool = MaskingTool(da)
    _, _, h = masking_tool.controls

    h.value = True
    h._tool.click(0.0, -30.0)
    h._tool.click(0.0, 1.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 1
    mask = next(iter(masks.values()))
    assert mask["kind"] == "hspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['y']["min"] == {"value": -30.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 1.0, "unit": "m"}

    h._tool.click(0.0, -10.0)
    h._tool.click(0.0, 17.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 2
    mask = list(masks.values())[1]
    assert mask["kind"] == "hspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['y']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 17.0, "unit": "m"}


@pytest.mark.usefixtures('_use_ipympl')
def test_mask_2d_rectangles_and_spans():
    da = make_data()
    masking_tool = MaskingTool(da)
    r, v, h = masking_tool.controls

    r.value = True
    r._tool.click(-10.0, -10.0)
    r._tool.click(5.0, 6.0)

    v.value = True
    v._tool.click(-20.0, 0.0)
    v._tool.click(1.0, 0.0)

    h.value = True
    h._tool.click(0.0, -30.0)
    h._tool.click(0.0, 1.0)

    masks = list(masking_tool.get_masks().values())
    assert len(masks) == 3

    mask = masks[0]
    assert mask["kind"] == "rectangle"
    bounds = mask["bounds"]
    assert len(bounds) == 2  # 2 dimensions
    assert bounds['x']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 5.0, "unit": "m"}
    assert bounds['y']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 6.0, "unit": "m"}

    mask = masks[1]
    assert mask["kind"] == "vspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['x']["min"] == {"value": -20.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 1.0, "unit": "m"}

    mask = masks[2]
    assert mask["kind"] == "hspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['y']["min"] == {"value": -30.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 1.0, "unit": "m"}


@pytest.mark.usefixtures('_use_ipympl')
def test_mask_1d():
    da = make_data().sum('y')
    masking_tool = MaskingTool(da)
    _, v, _ = masking_tool.controls

    v.value = True
    v._tool.click(-20.0, 0.0)
    v._tool.click(1.0, 0.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 1
    mask = next(iter(masks.values()))
    assert mask["kind"] == "vspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['x']["min"] == {"value": -20.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 1.0, "unit": "m"}

    v._tool.click(21.0, 0.0)
    v._tool.click(27.0, 0.0)

    masks = masking_tool.get_masks()
    assert len(masks) == 2
    mask = list(masks.values())[1]
    assert mask["kind"] == "vspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['x']["min"] == {"value": 21.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 27.0, "unit": "m"}


@pytest.mark.usefixtures('_use_ipympl')
def test_mask_color():
    da = make_data()
    color = to_hex('red')
    masking_tool = MaskingTool(da, color=color)
    r, v, _ = masking_tool.controls

    r.value = True
    r._tool.click(-10.0, -10.0)
    r._tool.click(5.0, 6.0)

    shape = r._tool.children[0]
    assert to_hex(shape.facecolor) == color
    assert to_hex(shape.edgecolor) == color

    v.value = True
    v._tool.click(-20.0, 0.0)
    v._tool.click(1.0, 0.0)

    shape = v._tool.children[0]
    assert to_hex(shape.facecolor) == color
    assert to_hex(shape.edgecolor) == color


@pytest.mark.usefixtures('_use_ipympl')
def test_save_masks(tmp_path):
    da = make_data()
    masking_tool = MaskingTool(da)
    r, v, h = masking_tool.controls

    r.value = True
    r._tool.click(-10.0, -10.0)
    r._tool.click(5.0, 6.0)

    v.value = True
    v._tool.click(-20.0, 0.0)
    v._tool.click(1.0, 0.0)

    h.value = True
    h._tool.click(0.0, -30.0)
    h._tool.click(0.0, 1.0)

    file = tmp_path / ".json"
    masking_tool.save_masks(file)

    with open(file) as f:
        loaded_masks = json.load(f)

    assert len(loaded_masks) == 3

    masks = list(loaded_masks.values())

    mask = masks[0]
    assert mask["kind"] == "rectangle"
    bounds = mask["bounds"]
    assert len(bounds) == 2  # 2 dimensions
    assert bounds['x']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 5.0, "unit": "m"}
    assert bounds['y']["min"] == {"value": -10.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 6.0, "unit": "m"}

    mask = masks[1]
    assert mask["kind"] == "vspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['x']["min"] == {"value": -20.0, "unit": "m"}
    assert bounds['x']["max"] == {"value": 1.0, "unit": "m"}

    mask = masks[2]
    assert mask["kind"] == "hspan"
    bounds = mask["bounds"]
    assert len(bounds) == 1  # 1 dimension
    assert bounds['y']["min"] == {"value": -30.0, "unit": "m"}
    assert bounds['y']["max"] == {"value": 1.0, "unit": "m"}


@pytest.mark.usefixtures('_use_ipympl')
def test_masking_tool_raises_with_3d_data():
    da = make_data() * sc.arange('z', 3)
    with pytest.raises(
        ValueError, match="The masking tool can only handle 1D or 2D data."
    ):
        MaskingTool(da)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import numpy as np
import pytest
import scipp as sc
import scipp.constants as const
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st
from hypothesis.extra import numpy as npst
from scipp.testing import strategies as scst

from scippneutron.conversion import beamline as beamline_conv
from scippneutron.conversion import tof as tof_conv

global_settings = {
    'max_examples': 20,
    # Generating variables needs a lot of memory. Nothing we can fix in scippneutron.
    'suppress_health_check': [HealthCheck.data_too_large],
}


def element_args():
    return {
        'allow_nan': False,
        'allow_infinity': False,
        'allow_subnormal': False,
        'min_value': 1e-2,
        'max_value': 1e6,
    }


def simple_variables(dims, unit):
    return scst.variables(
        sizes=st.dictionaries(
            keys=dims,
            values=st.integers(min_value=1, max_value=4),
            min_size=1,
            max_size=1,
        ),
        unit=unit,
        with_variances=False,
        dtype='float',
        elements=element_args(),
    )


def n_simple_variables(n, dims, unit):
    return scst.n_variables(
        n,
        sizes=st.dictionaries(
            keys=dims,
            values=st.integers(min_value=1, max_value=4),
            min_size=1,
            max_size=1,
        ),
        unit=unit,
        with_variances=False,
        dtype='float',
        elements=element_args(),
    )


def time_variables():
    return simple_variables(
        dims=st.sampled_from(('time', 't', 'tof')),
        unit=st.sampled_from(('s', 'ms', 'us')),
    )


def space_variables():
    return simple_variables(
        dims=st.sampled_from(('position', 'spectrum', 'x')),
        unit=st.sampled_from(('m', 'cm', 'angstrom')),
    )


def n_space_variables(n):
    return n_simple_variables(
        n,
        dims=st.sampled_from(('position', 'spectrum', 'x')),
        unit=st.sampled_from(('m', 'cm', 'angstrom')),
    )


def angle_variables():
    return simple_variables(
        dims=st.sampled_from(('angle', 'two_theta', 'alpha')),
        unit=st.sampled_from(('deg', 'rad')),
    )


def energy_variables():
    return simple_variables(
        dims=st.sampled_from(('energy', 'E', 'energy_transfer')),
        unit=st.sampled_from(('meV', 'J')),
    )


def vectors_variables():
    return st.fixed_dictionaries(
        {
            'size': st.integers(min_value=1, max_value=4),
            'dim': st.sampled_from(('position', 'detector', 'pixel')),
            'unit': st.sampled_from(('m', 'cm')),
        }
    ).flatmap(
        lambda args: npst.arrays(float, (args['size'], 3), elements=element_args()).map(
            lambda values: sc.vectors(
                dims=[args['dim']], values=values, unit=args['unit']
            )
        )
    )


def vector_variables():
    return st.sampled_from(('m', 'cm')).flatmap(
        lambda unit: npst.arrays(float, 3, elements=element_args()).map(
            lambda values: sc.vector(values, unit=unit)
        )
    )


@given(tof=time_variables(), Ltotal=space_variables())
@settings(**global_settings)
def test_wavelength_from_tof(tof, Ltotal):
    wavelength = tof_conv.wavelength_from_tof(tof=tof, Ltotal=Ltotal)
    assert sc.allclose(
        wavelength, sc.to_unit(const.h * tof / const.m_n / Ltotal, unit='angstrom')
    )


@pytest.mark.parametrize('tof_dtype', ['float64', 'int64', 'int32'])
@pytest.mark.parametrize('Ltotal_dtype', ['float64', 'float32', 'int64', 'int32'])
def test_wavelength_from_tof_double_precision(tof_dtype, Ltotal_dtype):
    tof = sc.scalar(1.2, unit='s', dtype=tof_dtype)
    Ltotal = sc.scalar(10.1, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.wavelength_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float64'


@pytest.mark.parametrize('Ltotal_dtype', ['float64', 'float32', 'int64', 'int32'])
def test_wavelength_from_tof_single_precision(Ltotal_dtype):
    tof = sc.scalar(1.2, unit='s', dtype='float32')
    Ltotal = sc.scalar(10.1, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.wavelength_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float32'


@given(
    tof=time_variables(),
    Ltotal_and_two_theta=n_space_variables(2),
    two_theta_unit=st.sampled_from(('deg', 'rad')),
)
@settings(**global_settings)
def test_dspacing_from_tof(tof, Ltotal_and_two_theta, two_theta_unit):
    Ltotal, two_theta = Ltotal_and_two_theta
    two_theta.unit = two_theta_unit
    dspacing = tof_conv.dspacing_from_tof(tof=tof, Ltotal=Ltotal, two_theta=two_theta)
    assert sc.allclose(
        dspacing,
        sc.to_unit(
            const.h * tof / const.m_n / Ltotal / 2 / sc.sin(two_theta / 2),
            unit='angstrom',
        ),
    )


@pytest.mark.parametrize('tof_dtype', ['float64', 'int64', 'int32'])
@pytest.mark.parametrize('Ltotal_dtype', ['float64', 'float32', 'int64', 'int32'])
@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64', 'int32'])
def test_dspacing_from_tof_double_precision(tof_dtype, Ltotal_dtype, two_theta_dtype):
    tof = sc.scalar(52.0, unit='s', dtype=tof_dtype)
    Ltotal = sc.scalar(0.341, unit='m', dtype=Ltotal_dtype)
    two_theta = sc.scalar(1.68, unit='rad', dtype=two_theta_dtype)
    assert (
        tof_conv.dspacing_from_tof(tof=tof, Ltotal=Ltotal, two_theta=two_theta).dtype
        == 'float64'
    )


@pytest.mark.parametrize('Ltotal_dtype', ['float64', 'float32', 'int64', 'int32'])
@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64', 'int32'])
def test_dspacing_from_tof_single_precision(Ltotal_dtype, two_theta_dtype):
    tof = sc.scalar(52.0, unit='s', dtype='float32')
    Ltotal = sc.scalar(0.341, unit='m', dtype=Ltotal_dtype)
    two_theta = sc.scalar(1.68, unit='rad', dtype=two_theta_dtype)
    assert (
        tof_conv.dspacing_from_tof(tof=tof, Ltotal=Ltotal, two_theta=two_theta).dtype
        == 'float32'
    )


@given(tof=time_variables(), Ltotal=space_variables())
@settings(**global_settings)
def test_energy_from_tof(tof, Ltotal):
    energy = tof_conv.energy_from_tof(tof=tof, Ltotal=Ltotal)
    assert sc.allclose(
        energy, sc.to_unit(const.m_n * Ltotal**2 / 2 / tof**2, unit='meV')
    )


@pytest.mark.parametrize('tof_dtype', ['float64', 'int64'])
@pytest.mark.parametrize('Ltotal_dtype', ['float64', 'float32', 'int64'])
def test_energy_from_tof_double_precision(tof_dtype, Ltotal_dtype):
    tof = sc.scalar(478.9, unit='s', dtype=tof_dtype)
    Ltotal = sc.scalar(1.256, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.energy_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float64'


@pytest.mark.parametrize('Ltotal_dtype', ['float64', 'float32', 'int64'])
def test_energy_from_tof_single_precision(Ltotal_dtype):
    tof = sc.scalar(478.9, unit='s', dtype='float32')
    Ltotal = sc.scalar(1.256, unit='m', dtype=Ltotal_dtype)
    assert tof_conv.energy_from_tof(tof=tof, Ltotal=Ltotal).dtype == 'float32'


@given(
    tof=time_variables(),
    L1_and_L2=n_space_variables(2),
    L2_unit=st.sampled_from(('m', 'mm', 'cm')),
    incident_energy=energy_variables(),
)
@settings(**global_settings)
def test_energy_transfer_direct_from_tof(tof, L1_and_L2, L2_unit, incident_energy):
    L1, L2 = L1_and_L2
    L2.unit = L2_unit
    # Energies are always > 0. This matters for detection of invalid tof values.
    incident_energy = abs(incident_energy) * 1.0001

    energy_transfer = tof_conv.energy_transfer_direct_from_tof(
        tof=tof, L1=L1, L2=L2, incident_energy=incident_energy
    )

    t0 = sc.to_unit(sc.sqrt(const.m_n * L1**2 / 2 / incident_energy), tof.unit)
    expected = incident_energy - sc.to_unit(
        const.m_n * L2**2 / 2 / (tof - t0) ** 2, incident_energy.unit
    )
    expected = sc.where(
        tof >= t0, expected, sc.scalar(np.nan, unit=incident_energy.unit)
    )
    assert sc.allclose(energy_transfer, expected, equal_nan=True)


@given(
    tof=time_variables(),
    L1_and_L2=n_space_variables(2),
    L2_unit=st.sampled_from(('m', 'mm', 'cm')),
    final_energy=energy_variables(),
)
@settings(**global_settings)
def test_energy_transfer_indirect_from_tof(tof, L1_and_L2, L2_unit, final_energy):
    L1, L2 = L1_and_L2
    L2.unit = L2_unit
    # Energies are always > 0. This matters for detection of invalid tof values.
    final_energy = abs(final_energy) * 1.0001

    energy_transfer = tof_conv.energy_transfer_indirect_from_tof(
        tof=tof, L1=L1, L2=L2, final_energy=final_energy
    )

    t0 = sc.to_unit(sc.sqrt(const.m_n * L2**2 / 2 / final_energy), tof.unit)
    expected = (
        sc.to_unit(const.m_n * L1**2 / 2 / (tof - t0) ** 2, final_energy.unit)
        - final_energy
    )
    expected = sc.where(tof >= t0, expected, sc.scalar(np.nan, unit=final_energy.unit))
    assert sc.allclose(energy_transfer, expected, equal_nan=True)


@given(wavelength=space_variables())
@settings(**global_settings)
def test_energy_from_wavelength(wavelength):
    energy = tof_conv.energy_from_wavelength(wavelength=wavelength)
    assert sc.allclose(
        energy, sc.to_unit(const.h**2 / 2 / const.m_n / wavelength**2, unit='meV')
    )


@pytest.mark.parametrize('wavelength_dtype', ['float64', 'int64'])
def test_energy_from_wavelength_double_precision(wavelength_dtype):
    wavelength = sc.scalar(60.5, unit='m', dtype=wavelength_dtype)
    assert tof_conv.energy_from_wavelength(wavelength=wavelength).dtype == 'float64'


def test_energy_from_wavelength_single_precision():
    wavelength = sc.scalar(60.5, unit='m', dtype='float32')
    assert tof_conv.energy_from_wavelength(wavelength=wavelength).dtype == 'float32'


@given(energy=energy_variables())
@settings(**global_settings)
def test_wavelength_from_energy(energy):
    wavelength = tof_conv.wavelength_from_energy(energy=energy)
    assert sc.allclose(
        wavelength,
        sc.to_unit(const.h / sc.sqrt(2 * const.m_n * energy), unit='angstrom'),
    )


@pytest.mark.parametrize('energy_dtype', ['float64', 'int64'])
def test_wavelength_from_energy_double_precision(energy_dtype):
    energy = sc.scalar(61.0, unit='meV', dtype=energy_dtype)
    assert tof_conv.wavelength_from_energy(energy=energy).dtype == 'float64'


def test_wavelength_from_energy_single_precision():
    energy = sc.scalar(61.0, unit='meV', dtype='float32')
    assert tof_conv.wavelength_from_energy(energy=energy).dtype == 'float32'


@given(wavelength=space_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_Q_from_wavelength(wavelength, two_theta):
    Q = tof_conv.Q_from_wavelength(wavelength=wavelength, two_theta=two_theta)
    assert sc.allclose(Q, 4 * np.pi * sc.sin(two_theta / 2) / wavelength)


@pytest.mark.parametrize('wavelength_dtype', ['float64', 'int64'])
@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_Q_from_wavelength_double_precision(wavelength_dtype, two_theta_dtype):
    wavelength = sc.scalar(3.51, unit='s', dtype=wavelength_dtype)
    two_theta = sc.scalar(0.041, unit='deg', dtype=two_theta_dtype)
    assert (
        tof_conv.Q_from_wavelength(wavelength=wavelength, two_theta=two_theta).dtype
        == 'float64'
    )


@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_Q_from_wavelength_single_precision(two_theta_dtype):
    wavelength = sc.scalar(3.51, unit='s', dtype='float32')
    two_theta = sc.scalar(0.041, unit='deg', dtype=two_theta_dtype)
    assert (
        tof_conv.Q_from_wavelength(wavelength=wavelength, two_theta=two_theta).dtype
        == 'float32'
    )


@given(Q=space_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_wavelength_from_Q(Q, two_theta):
    Q.unit = f'1/{Q.unit}'
    wavelength = tof_conv.wavelength_from_Q(Q=Q, two_theta=two_theta)
    assert sc.allclose(
        wavelength, sc.to_unit(4 * np.pi * sc.sin(two_theta / 2) / Q, unit='angstrom')
    )


@pytest.mark.parametrize('Q_dtype', ['float64', 'int64'])
@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_wavelength_from_Q_double_precision(Q_dtype, two_theta_dtype):
    Q = sc.scalar(4.151, unit='1/nm', dtype=Q_dtype)
    two_theta = sc.scalar(5.71, unit='deg', dtype=two_theta_dtype)
    assert tof_conv.wavelength_from_Q(Q=Q, two_theta=two_theta).dtype == 'float64'


@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_wavelength_from_Q_single_precision(two_theta_dtype):
    Q = sc.scalar(4.151, unit='1/nm', dtype='float32')
    two_theta = sc.scalar(5.71, unit='deg', dtype=two_theta_dtype)
    assert tof_conv.wavelength_from_Q(Q=Q, two_theta=two_theta).dtype == 'float32'


@given(incident_beam=vector_variables(), wavelength=space_variables())
@settings(**global_settings)
def test_Q_elements_from_wavelength_equal_beams(incident_beam, wavelength):
    scattered_beam = incident_beam.copy()
    (Qx, Qy, Qz) = tof_conv.Q_elements_from_wavelength(
        wavelength=wavelength,
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
    )
    assert sc.allclose(Qx, 0.0 / wavelength.unit)
    assert sc.allclose(Qy, 0.0 / wavelength.unit)
    assert sc.allclose(Qz, 0.0 / wavelength.unit)


@given(
    incident_beam=vector_variables(),
    scattered_beam=vectors_variables(),
    wavelength=space_variables(),
)
@settings(**global_settings)
def test_Q_elements_from_wavelength_consistent_with_Q(
    incident_beam, scattered_beam, wavelength
):
    wavelength = wavelength.rename({wavelength.dim: 'wavelength'})
    (Qx, Qy, Qz) = tof_conv.Q_elements_from_wavelength(
        wavelength=wavelength,
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
    )
    Q_vec = sc.spatial.as_vectors(Qx, Qy, Qz)
    Q = tof_conv.Q_from_wavelength(
        wavelength=wavelength,
        two_theta=beamline_conv.two_theta(
            incident_beam=incident_beam, scattered_beam=scattered_beam
        ),
    )
    assert sc.allclose(sc.norm(Q_vec), Q)


def test_Q_elements_from_wavelength():
    incident_beam = sc.vector([0.0, 0.0, 10.0], unit='m')
    scattered_beam = sc.vectors(
        dims=['pos'],
        values=[[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 2.0]],
        unit='cm',
    )
    wavelength = sc.array(dims=['wavelength'], values=[0.3, 2.0], unit='angstrom')
    (Qx, Qy, Qz) = tof_conv.Q_elements_from_wavelength(
        wavelength=wavelength,
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
    )
    Q_vec = sc.spatial.as_vectors(Qx, Qy, Qz)
    assert sc.identical(
        Q_vec['pos', 0], 2 * np.pi / wavelength * sc.vector([-1.0, 0.0, 1.0])
    )
    assert sc.identical(
        Q_vec['pos', 1], 2 * np.pi / wavelength * sc.vector([0.0, -1.0, 1.0])
    )
    assert sc.allclose(
        Q_vec['pos', 2],
        2
        * np.pi
        / wavelength
        * sc.vector([0.0, -0.4472135954999579, 0.1055728090000841]),
    )


@given(wavelength=space_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_dspacing_from_wavelength(wavelength, two_theta):
    dspacing = tof_conv.dspacing_from_wavelength(
        wavelength=wavelength, two_theta=two_theta
    )
    assert sc.allclose(
        dspacing, sc.to_unit(wavelength / 2 / sc.sin(two_theta / 2), unit='angstrom')
    )


@pytest.mark.parametrize('wavelength_dtype', ['float64', 'int64'])
@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_dspacing_from_wavelength_double_precision(wavelength_dtype, two_theta_dtype):
    wavelength = sc.scalar(41.4, unit='m', dtype=wavelength_dtype)
    two_theta = sc.scalar(8.4, unit='rad', dtype=two_theta_dtype)
    assert (
        tof_conv.dspacing_from_wavelength(
            wavelength=wavelength, two_theta=two_theta
        ).dtype
        == 'float64'
    )


@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_dspacing_from_wavelength_single_precision(two_theta_dtype):
    wavelength = sc.scalar(41.4, unit='m', dtype='float32')
    two_theta = sc.scalar(8.4, unit='rad', dtype=two_theta_dtype)
    assert (
        tof_conv.dspacing_from_wavelength(
            wavelength=wavelength, two_theta=two_theta
        ).dtype
        == 'float32'
    )


@given(energy=energy_variables(), two_theta=angle_variables())
@settings(**global_settings)
def test_dspacing_from_energy(energy, two_theta):
    dspacing = tof_conv.dspacing_from_energy(energy=energy, two_theta=two_theta)
    assert sc.allclose(
        dspacing,
        sc.to_unit(
            const.h / sc.sqrt(8 * const.m_n * energy) / sc.sin(two_theta / 2),
            unit='angstrom',
        ),
    )


@pytest.mark.parametrize('energy_dtype', ['float64', 'int64'])
@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_dspacing_from_energy_double_precision(energy_dtype, two_theta_dtype):
    energy = sc.scalar(26.90, unit='J', dtype=energy_dtype)
    two_theta = sc.scalar(1.985, unit='rad', dtype=two_theta_dtype)
    assert (
        tof_conv.dspacing_from_energy(energy=energy, two_theta=two_theta).dtype
        == 'float64'
    )


@pytest.mark.parametrize('two_theta_dtype', ['float64', 'float32', 'int64'])
def test_dspacing_from_energy_single_precision(two_theta_dtype):
    energy = sc.scalar(26.90, unit='J', dtype='float32')
    two_theta = sc.scalar(1.985, unit='rad', dtype=two_theta_dtype)
    assert (
        tof_conv.dspacing_from_energy(energy=energy, two_theta=two_theta).dtype
        == 'float32'
    )


def test_q_vec_from_q_elements():
    Qx, Qy, Qz = (
        sc.array(dims=['Q'], values=[2.1 * i, i / 3], unit='1/angstrom')
        for i in range(3)
    )
    q_vec = tof_conv.Q_vec_from_Q_elements(Qx=Qx, Qy=Qy, Qz=Qz)
    sc.testing.assert_identical(q_vec.fields.x, Qx)
    sc.testing.assert_identical(q_vec.fields.y, Qy)
    sc.testing.assert_identical(q_vec.fields.z, Qz)


def test_q_vec_from_q_elements_raises_for_shape_mismatch():
    Qx = sc.array(dims=['q'], values=[2.0, 3.0])
    Qy = Qx.copy()
    Qz = sc.array(dims=['q'], values=[2.0, 3.0, 4.0])
    with pytest.raises(sc.DimensionError):
        tof_conv.Q_vec_from_Q_elements(Qx=Qx, Qy=Qy, Qz=Qz)


def test_q_vec_from_q_elements_raises_for_dim_mismatch():
    Qx = sc.array(dims=['q'], values=[2.0, 3.0])
    Qy = Qx.copy()
    Qz = sc.array(dims=['Q'], values=[2.0, 3.0])
    with pytest.raises(sc.DimensionError):
        tof_conv.Q_vec_from_Q_elements(Qx=Qx, Qy=Qy, Qz=Qz)


def test_q_vec_from_q_elements_raises_for_unit_mismatch():
    Qx, Qy, Qz = (
        sc.array(dims=['Q'], values=[2.1 * i, i / 3], unit='1/angstrom')
        for i in range(3)
    )
    Qy.unit = '1/m'
    with pytest.raises(sc.UnitError):
        tof_conv.Q_vec_from_Q_elements(Qx=Qx, Qy=Qy, Qz=Qz)


def make_b_matrix() -> sc.Variable:
    """Return a B matrix.

    Uses and Levi convention
      W. R. Busing and H. A. Levy
      Angle calculations for 3- and 4-circle X-ray and neutron diffractometers
      Acta Cryst. (1967). 22, 457-464)
    """
    a = 3.1
    b = 0.3
    c = 1.1
    alpha = 0.7
    beta = 4.6
    gamma = 1.4

    return sc.spatial.linear_transform(
        value=[
            [a, b * np.cos(gamma), c * np.cos(beta)],
            [0, b * np.sin(gamma), -c * np.sin(beta) * np.cos(alpha)],
            [0, 0, 1 / c],
        ],
        unit='1/angstrom',
    )


def test_ub_matrix_from_u_and_b():
    b_matrix = make_b_matrix()
    u_coeffs = np.array([0.5, 0.1, -0.4, 0.9])
    u_matrix = sc.spatial.rotation(value=u_coeffs / np.linalg.norm(u_coeffs))
    expected = u_matrix * b_matrix

    assert sc.allclose(
        tof_conv.ub_matrix_from_u_and_b(u_matrix=u_matrix, b_matrix=b_matrix), expected
    )


@given(inv_q=n_space_variables(3))
def test_hkl_vec_from_Q_vec(inv_q):
    Qx, Qy, Qz = (sc.reciprocal(x.to(dtype='float64')) for x in inv_q)
    Q_vec = tof_conv.Q_vec_from_Q_elements(Qx=Qx, Qy=Qy, Qz=Qz)

    b_matrix = make_b_matrix()
    u_coeffs = np.array([0.5, 0.1, -0.4, 0.9])
    u_matrix = sc.spatial.rotation(value=u_coeffs / np.linalg.norm(u_coeffs))
    ub_matrix = u_matrix * b_matrix

    sample_rotation_coeffs = np.array([1.2, 0.8, -1.2, -0.1])
    sample_rotation_matrix = sc.spatial.rotation(
        value=sample_rotation_coeffs / np.linalg.norm(sample_rotation_coeffs)
    )

    hkl_vec = tof_conv.hkl_vec_from_Q_vec(
        Q_vec=Q_vec, ub_matrix=ub_matrix, sample_rotation=sample_rotation_matrix
    )

    # hkl is dimensionless but may have a multiplier.
    assert 'powers' not in hkl_vec.unit.to_dict()

    assert hkl_vec.sizes == Q_vec.sizes
    assert hkl_vec.dtype == sc.DType.vector3

    reconstructed_Q = (
        2 * np.pi * (sample_rotation_matrix * u_matrix * b_matrix * hkl_vec)
    )
    assert sc.allclose(reconstructed_Q, Q_vec)


def test_hkl_elements_from_hkl_vec():
    hkl_vec = sc.vector([3.1, 4.5, 6.9])
    h, k, l = tof_conv.hkl_elements_from_hkl_vec(hkl_vec=hkl_vec)  # noqa: E741
    sc.testing.assert_identical(h, hkl_vec.fields.x)
    sc.testing.assert_identical(k, hkl_vec.fields.y)
    sc.testing.assert_identical(l, hkl_vec.fields.z)


@pytest.mark.parametrize('pulse_time', [0.0, 1.0])
def test_time_at_sample(pulse_time):
    ts = tof_conv.time_at_sample_from_tof(
        pulse_time=sc.scalar(pulse_time, unit='s'),
        tof=sc.scalar(3.0, unit='s'),
        L2=sc.scalar(2.0, unit='m'),
        wavelength=(sc.constants.h / sc.constants.m_n / sc.scalar(2.0, unit='m/s')).to(
            unit='Å'
        ),
    )
    assert sc.allclose(ts, sc.scalar(pulse_time + 2.0, unit='s'))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import pytest

from scippneutron.conversion.graph import tof


def test_elastic():
    assert set(tof.elastic('energy').keys()) == {'dspacing', 'wavelength'}
    assert set(tof.elastic('tof').keys()) == {
        'dspacing',
        'energy',
        'Q',
        'wavelength',
        ('Qx', 'Qy', 'Qz'),
        'Q_vec',
        'hkl_vec',
        ('h', 'k', 'l'),
        'ub_matrix',
        'time_at_sample',
    }
    assert set(tof.elastic('Q').keys()) == {'wavelength'}
    assert set(tof.elastic('wavelength').keys()) == {
        'dspacing',
        'energy',
        'Q',
        ('Qx', 'Qy', 'Qz'),
        'Q_vec',
        'hkl_vec',
        ('h', 'k', 'l'),
        'ub_matrix',
    }


@pytest.mark.parametrize('start', ['dspacing'])
def test_elastic_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.elastic(start)


def test_kinematic():
    assert set(tof.kinematic('tof').keys()) == {'energy', 'wavelength'}
    assert set(tof.kinematic('wavelength').keys()) == {'energy'}
    assert set(tof.kinematic('energy').keys()) == {'wavelength'}


@pytest.mark.parametrize('start', ['dspacing', 'Q'])
def test_kinematic_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.kinematic(start)


def test_elastic_dspacing():
    assert set(tof.elastic_dspacing('energy').keys()) == {'dspacing'}
    assert set(tof.elastic_dspacing('tof').keys()) == {'dspacing'}
    assert set(tof.elastic_dspacing('wavelength').keys()) == {'dspacing'}


@pytest.mark.parametrize('start', ['Q'])
def test_elastic_dspacing_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.elastic_dspacing(start)


def test_elastic_energy():
    assert set(tof.elastic_energy('tof').keys()) == {'energy'}
    assert set(tof.elastic_energy('wavelength').keys()) == {'energy'}


@pytest.mark.parametrize('start', ['Q', 'dspacing'])
def test_elastic_energy_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.elastic_energy(start)


def test_elastic_Q():
    assert set(tof.elastic_Q('tof').keys()) == {'Q', 'wavelength'}
    assert set(tof.elastic_Q('wavelength').keys()) == {'Q'}


@pytest.mark.parametrize('start', ['energy', 'dspacing'])
def test_elastic_Q_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.elastic_Q(start)


def test_elastic_wavelength():
    assert set(tof.elastic_wavelength('energy').keys()) == {'wavelength'}
    assert set(tof.elastic_wavelength('tof').keys()) == {'wavelength'}
    assert set(tof.elastic_wavelength('Q').keys()) == {'wavelength'}


@pytest.mark.parametrize('start', ['dspacing'])
def test_elastic_wavelength_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.elastic_wavelength(start)


def test_direct_inelastic():
    assert set(tof.direct_inelastic('tof').keys()) == {'energy_transfer'}


@pytest.mark.parametrize('start', ['wavelength', 'Q', 'dspacing'])
def test_direct_inelastic_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.direct_inelastic(start)


def test_indirect_inelastic():
    assert set(tof.indirect_inelastic('tof').keys()) == {'energy_transfer'}


@pytest.mark.parametrize('start', ['wavelength', 'Q', 'dspacing'])
def test_indirect_inelastic_unsupported_starts(start):
    with pytest.raises(KeyError):
        tof.indirect_inelastic(start)


@pytest.mark.parametrize(
    'arg',
    [
        (tof.elastic, ('energy', 'tof', 'Q', 'wavelength')),
        (tof.kinematic, ('tof', 'wavelength', 'energy')),
        (tof.elastic_dspacing, ('tof', 'wavelength', 'energy')),
        (tof.elastic_energy, ('tof', 'wavelength')),
        (tof.elastic_Q, ('tof', 'wavelength')),
        (tof.elastic_wavelength, ('tof', 'energy', 'Q')),
        (tof.direct_inelastic, ('tof',)),
        (tof.indirect_inelastic, ('tof',)),
    ],
)
def test_returns_new_graph(arg):
    fn, starts = arg
    for start in starts:
        g = fn(start)
        g['a_new_node'] = lambda position: 2 * position
        assert 'a_new_node' not in fn(start)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from ._disk_chopper_repr import disk_chopper_html_repr

__all__ = ['disk_chopper_html_repr']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

"""Components for coordinate transformations.

The submodules are mainly intended for constructing graphs for
:func:`scipp.transform_coords` but can also be used independently.

``beamline`` and ``tof`` contain individual transformation kernels.
``graph`` contains concrete graphs that can be passed to ``transform_coords``.

See also the user guide on
`Coordinate Transformations <../../user-guide/coordinate-transformations.rst>`_.
"""

from . import beamline, tof

__all__ = ['beamline', 'tof']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

import numpy as np
import pytest
import scipp as sc
import scipp.constants
import scipp.testing
from hypothesis import given, settings
from hypothesis import strategies as st
from hypothesis.extra import numpy as npst

from scippneutron.conversion import beamline


def rotations() -> st.SearchStrategy[sc.Variable]:
    coefficients = st.floats(
        allow_nan=False, allow_infinity=False, allow_subnormal=False
    )
    return (
        npst.arrays(dtype=np.float64, shape=4, elements=coefficients)
        .filter(lambda q: np.linalg.norm(q) != 0)
        .map(lambda q: sc.spatial.rotation(value=q / np.linalg.norm(q)))
    )


def test_straight_incident_beam():
    source_position = sc.vectors(
        dims=['siti'], values=[[40, 80, 20], [30, 10, 50]], unit='mm'
    )
    sample_position = sc.vector([7, 5, 3], unit='mm')
    incident_beam = beamline.straight_incident_beam(
        source_position=source_position, sample_position=sample_position
    )
    sc.testing.assert_allclose(
        incident_beam,
        sc.vectors(dims=['siti'], values=[[-33, -75, -17], [-23, -5, -47]], unit='mm'),
    )


def test_straight_scattered_beam():
    position = sc.vectors(dims=['on'], values=[[6, 3, 9], [8, 3, 1]], unit='km')
    sample_position = sc.vector([0.4, 0.6, 0.2], unit='km')
    scattered_beam = beamline.straight_scattered_beam(
        position=position, sample_position=sample_position
    )
    sc.testing.assert_allclose(
        scattered_beam,
        sc.vectors(dims=['on'], values=[[5.6, 2.4, 8.8], [7.6, 2.4, 0.8]], unit='km'),
    )


def test_L1():
    incident_beam = sc.vectors(
        dims=['inc'], values=[[0.5, 1.0, 1.5], [-0.3, 0.6, -0.9]], unit='um'
    )
    L1 = beamline.L1(incident_beam=incident_beam)
    sc.testing.assert_allclose(
        L1, sc.array(dims=['inc'], values=[1.870828693386, 1.122497216032], unit='um')
    )


def test_L2():
    scattered_beam = sc.vectors(
        dims=['scat'], values=[[11, 22, 33], [95, 84, 73]], unit='am'
    )
    L2 = beamline.L2(scattered_beam=scattered_beam)
    sc.testing.assert_allclose(
        L2,
        sc.array(dims=['scat'], values=[41.158231254513, 146.321563687653], unit='am'),
    )


def test_total_beam_length():
    L1 = sc.scalar(5.134, unit='cm')
    L2 = sc.array(dims=['secondary'], values=[3.14159, 42.0, 999.0], unit='cm')
    Ltotal = beamline.total_beam_length(L1=L1, L2=L2)
    sc.testing.assert_allclose(
        Ltotal,
        sc.array(dims=['secondary'], values=[8.27559, 47.134, 1004.134], unit='cm'),
    )


def test_total_straight_beam_length_no_scatter():
    position = sc.vectors(dims=['po'], values=[[1, 2, 3], [4, 5, 6]], unit='m')
    source_position = sc.vector([0.1, 0.2, 0.3], unit='m')
    Ltotal = beamline.total_straight_beam_length_no_scatter(
        source_position=source_position, position=position
    )
    sc.testing.assert_allclose(
        Ltotal, sc.array(dims=['po'], values=[3.367491648096, 8.410707461325], unit='m')
    )


def test_two_theta_arbitrary_values():
    incident_beam = sc.vector([0.564, 1.2, -10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[13, 24, 35], [51, -42, 33]], unit='m'
    )
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    sc.testing.assert_allclose(
        two_theta,
        sc.array(dims=['beam'], values=[2.352629742382, 2.061447408052], unit='rad'),
    )


def test_two_theta_depends_on_beam_direction():
    incident_beam = sc.vector([0.564, 1.2, -10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[13, 24, 35], [51, -42, 33]], unit='m'
    )
    a = beamline.two_theta(incident_beam=incident_beam, scattered_beam=scattered_beam)
    b = beamline.two_theta(incident_beam=-incident_beam, scattered_beam=scattered_beam)
    sc.testing.assert_allclose(
        a,
        sc.scalar(np.pi, unit='rad') - b,
    )


def test_two_theta_parallel_beams():
    incident_beam = sc.vector([1.0, 0.2, 0.0], unit='mm')
    scattered_beam = sc.array(dims=['beam'], values=[1.0, 2.0, 0.1]) * incident_beam
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    sc.testing.assert_allclose(
        two_theta,
        sc.zeros(dims=['beam'], shape=[3], unit='rad'),
        atol=sc.scalar(1e-14, unit='rad'),
    )


def test_two_theta_anti_parallel_beams():
    incident_beam = sc.vector([10.2, -0.8, 4.1], unit='mm')
    scattered_beam = sc.array(dims=['beam'], values=[-2.1, -31.0, -1.0]) * incident_beam
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    sc.testing.assert_allclose(
        two_theta, sc.full(value=np.pi, dims=['beam'], shape=[3], unit='rad')
    )


def test_two_theta_orthogonal_beams():
    incident_beam = sc.vector([0.1, 0.0, 10.0], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[0.0, -2.1, 0.0], [10.0, 123.0, -0.1]], unit='m'
    )
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    sc.testing.assert_allclose(
        two_theta, sc.full(value=np.pi / 2, dims=['beam'], shape=[2], unit='rad')
    )


def test_two_theta_invariant_under_reflection_about_incident_beam():
    incident_beam = sc.vector([0.564, 1.2, -10.4], unit='cm')
    rotation = sc.spatial.rotations_from_rotvecs(
        sc.vectors(
            dims=['scattered'], values=[[0.7, 0.0, 0.0], [-0.7, 0.0, 0.0]], unit='rad'
        )
    )
    scattered_beam = rotation * incident_beam
    two_theta = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )
    sc.testing.assert_allclose(two_theta[0], two_theta[1])


@given(rotation=rotations())
@settings(max_examples=10)
def test_two_theta_invariant_under_rotation(rotation: sc.Variable) -> None:
    incident_beam = sc.vector([0.564, 1.2, -10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[13, 24, 35], [51, -42, 33]], unit='m'
    )
    original = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )

    rotated = beamline.two_theta(
        incident_beam=rotation * incident_beam, scattered_beam=rotation * scattered_beam
    )

    sc.testing.assert_allclose(original, rotated)


@given(incident_rotation=rotations(), scattered_rotation=rotations())
@settings(max_examples=50)
def test_scattering_angles_with_gravity_small_gravity(
    incident_rotation: sc.Variable, scattered_rotation: sc.Variable
) -> None:
    # This case is unphysical but tests the consistency with `two_theta`.
    incident_beam = sc.vector([0.564, 0.0, 10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'],
        values=[[13, 24, 35], [51, -42, 33], [4, 23, -17], [-19, -31, 5]],
        unit='m',
    )
    incident_beam = incident_rotation * incident_beam
    scattered_beam = scattered_rotation * scattered_beam
    wavelength = sc.array(dims=['wavelength'], values=[1.2, 1.6, 1.8], unit='Å')
    gravity = sc.vector([0, -1e-11, 0], unit=sc.constants.g.unit)

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    ).broadcast(dims=['wavelength', 'beam'], shape=[3, 4])
    sc.testing.assert_allclose(res['two_theta'], expected)


def test_scattering_angles_with_gravity_small_gravity_orthogonal_coords() -> None:
    # This case is unphysical but tests the consistency with `two_theta`.
    # Here, incident_beam and gravity are orthogonal to use the optimized code.
    incident_beam = sc.vector([0.564, 0.0, 10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'],
        values=[[13, 24, 35], [51, -42, 33], [4, 23, -17], [-19, -31, 5]],
        unit='m',
    )
    wavelength = sc.array(dims=['wavelength'], values=[1.2, 1.6, 1.8], unit='Å')
    gravity = sc.vector([0, -1e-11, 0], unit=sc.constants.g.unit)

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    ).broadcast(dims=['wavelength', 'beam'], shape=[3, 4])
    sc.testing.assert_allclose(res['two_theta'], expected)


@pytest.mark.parametrize('polar', [np.pi / 3, np.pi / 2, 2 * np.pi / 3, np.pi])
@pytest.mark.parametrize('azimuthal', [0.0, np.pi / 2, np.pi])
def test_scattering_angles_with_gravity_reproduces_angles(
    polar: float, azimuthal: float
):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.

    gravity = sc.vector([1e-12, -1e-11, 2e-12], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot1 = sc.spatial.rotations_from_rotvecs(sc.vector([-polar, 0, 0], unit='rad'))
    rot2 = sc.spatial.rotations_from_rotvecs(
        sc.vector([0, 0, azimuthal - np.pi / 2], unit='rad')
    )
    scattered_beam = rot2 * (rot1 * incident_beam)

    wavelength = sc.scalar(1e-6, unit='Å')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'],
        sc.scalar(polar, unit='rad'),
        atol=sc.scalar(1e-15, unit='rad'),
    )
    # We can't easily test phi here because it depends on the rotation of gravity.


@pytest.mark.parametrize('polar', [np.pi / 3, np.pi / 2, 2 * np.pi / 3, np.pi])
@pytest.mark.parametrize('azimuthal', [0.0, np.pi / 2, np.pi])
def test_scattering_angles_with_gravity_reproduces_angles_orthogonal_coords(
    polar: float,
    azimuthal: float,
):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.
    # Here, incident_beam and gravity are orthogonal to use the optimized code.

    gravity = sc.vector([0.0, -1e-11, 0.0], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot1 = sc.spatial.rotations_from_rotvecs(sc.vector([-polar, 0, 0], unit='rad'))
    rot2 = sc.spatial.rotations_from_rotvecs(
        sc.vector([0, 0, azimuthal - np.pi / 2], unit='rad')
    )
    scattered_beam = rot2 * (rot1 * incident_beam)

    wavelength = sc.scalar(1e-6, unit='Å')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'],
        sc.scalar(polar, unit='rad'),
        atol=sc.scalar(1e-15, unit='rad'),
    )
    sc.testing.assert_allclose(
        res['phi'], sc.scalar(azimuthal, unit='rad'), atol=sc.scalar(1e-15, unit='rad')
    )


@pytest.mark.parametrize('polar', [np.pi / 3, np.pi / 2, 2 * np.pi / 3, np.pi])
@pytest.mark.parametrize('azimuthal', [3 * np.pi / 2, 8 * np.pi / 5, 2 * np.pi])
def test_scattering_angles_with_gravity_reproduces_angles_azimuth_greater_pi(
    polar: float, azimuthal: float
):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.

    gravity = sc.vector([-2e-12, -1e-11, 1e-13], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot1 = sc.spatial.rotations_from_rotvecs(sc.vector([-polar, 0, 0], unit='rad'))
    rot2 = sc.spatial.rotations_from_rotvecs(
        sc.vector([0, 0, azimuthal - np.pi / 2], unit='rad')
    )
    scattered_beam = rot2 * (rot1 * incident_beam)

    wavelength = sc.scalar(1e-6, unit='Å')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'],
        sc.scalar(polar, unit='rad'),
        atol=sc.scalar(1e-15, unit='rad'),
    )
    # We can't easily test phi here because it depends on the rotation of gravity.


@pytest.mark.parametrize('polar', [np.pi / 3, np.pi / 2, 2 * np.pi / 3, np.pi])
@pytest.mark.parametrize('azimuthal', [3 * np.pi / 2, 8 * np.pi / 5, 2 * np.pi])
def test_scattering_angles_with_gravity_reproduces_angles_azimuth_greater_pi_orthogonal_coords(  # noqa: E501
    polar: float, azimuthal: float
):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.
    # Here, incident_beam and gravity are orthogonal to use the optimized code.

    gravity = sc.vector([0.0, -1e-11, 0.0], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot1 = sc.spatial.rotations_from_rotvecs(sc.vector([-polar, 0, 0], unit='rad'))
    rot2 = sc.spatial.rotations_from_rotvecs(
        sc.vector([0, 0, azimuthal - np.pi / 2], unit='rad')
    )
    scattered_beam = rot2 * (rot1 * incident_beam)

    wavelength = sc.scalar(1e-6, unit='Å')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'],
        sc.scalar(polar, unit='rad'),
        atol=sc.scalar(1e-15, unit='rad'),
    )
    # phi has range (-pi, pi], so for azimuthal > pi, we get a negative result
    sc.testing.assert_allclose(
        res['phi'],
        sc.scalar(azimuthal - 2 * np.pi, unit='rad'),
        atol=sc.scalar(1e-15, unit='rad'),
    )


@pytest.mark.parametrize('azimuthal', [0.0, np.pi / 2, np.pi, 7 * np.pi / 5, 2 * np.pi])
def test_scattering_angles_with_gravity_reproduces_angles_polar_zero(azimuthal: float):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.

    gravity = sc.vector([0.0, -1e-11, 0.0], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot = sc.spatial.rotations_from_rotvecs(
        sc.vector([0, 0, azimuthal - np.pi / 2], unit='rad')
    )
    scattered_beam = rot * incident_beam

    wavelength = sc.scalar(1e-6, unit='Å')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'], sc.scalar(0.0, unit='rad'), atol=sc.scalar(1e-15, unit='rad')
    )
    # When polar = 0, the azimuthal angle is ill-defined, so there is no test.


def test_scattering_angles_with_gravity_drops_in_expected_direction():
    wavelength = sc.scalar(1.6, unit='Å')
    gravity = sc.vector([0.0, -sc.constants.g.value, 0.0], unit=sc.constants.g.unit)
    incident_beam = sc.vector([0.0, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[0.0, 2.5, 8.6], [0.0, -1.7, 6.9]], unit='m'
    )

    with_gravity = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    without_gravity = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )

    # The neutron was detected above the incident beam.
    # So using straight paths, it looks like it scattered at a
    # smaller angle (detected at a lower y) than in the real case with gravity.
    assert sc.all(with_gravity['two_theta'][0] > without_gravity[0]).value
    # The neutron was detected below the incident beam.
    # So the opposite of the above comment applies.
    assert sc.all(with_gravity['two_theta'][1] < without_gravity[1]).value
    sc.testing.assert_allclose(
        with_gravity['phi'],
        sc.array(dims=['det'], values=[np.pi / 2, -np.pi / 2], unit='rad'),
    )


def test_scattering_angles_with_gravity_beams_aligned_with_lab_coords():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    gravity = sc.vector([0.0, -sc.constants.g.value, 0.0], unit=sc.constants.g.unit)
    incident_beam = sc.vector([0.0, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[0.0, 2.5, 3.6], [0.0, -1.7, 2.9]], unit='m'
    )

    original_wavelength = wavelength.copy()
    original_gravity = gravity.copy()
    original_incident_beam = incident_beam.copy()
    original_scattered_beam = scattered_beam.copy()

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    L2 = sc.norm(scattered_beam)
    x = scattered_beam.fields.x
    y = scattered_beam.fields.y
    z = scattered_beam.fields.z
    drop = (
        L2**2
        * sc.norm(gravity)
        * wavelength**2
        * sc.constants.m_n**2
        / (2 * sc.constants.h**2)
    )
    drop = drop.to(unit=y.unit)
    true_y = y + drop
    true_scattered_beam = sc.spatial.as_vectors(x, true_y, z)
    expected_two_theta = sc.asin(
        sc.sqrt(x**2 + true_y**2) / sc.norm(true_scattered_beam)
    )
    expected_two_theta = expected_two_theta.transpose(res['two_theta'].dims)

    expected_phi = sc.atan2(y=true_y, x=x).transpose(res['phi'].dims)

    sc.testing.assert_allclose(res['two_theta'], expected_two_theta)
    sc.testing.assert_allclose(res['phi'], expected_phi)

    sc.testing.assert_identical(wavelength, original_wavelength)
    sc.testing.assert_identical(gravity, original_gravity)
    sc.testing.assert_identical(incident_beam, original_incident_beam)
    sc.testing.assert_identical(scattered_beam, original_scattered_beam)


def _reference_scattering_angles_with_gravity(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    gravity: sc.Variable,
    wavelength: sc.Variable,
) -> dict[str, sc.Variable]:
    # This is a simplified, independently checked implementation.
    e_z = incident_beam / sc.norm(incident_beam)
    e_y = -gravity / sc.norm(gravity)
    e_x = sc.cross(e_y, e_z)

    x = sc.dot(scattered_beam, e_x)
    y = sc.dot(scattered_beam, e_y)
    z = sc.dot(scattered_beam, e_z)

    L2 = sc.norm(scattered_beam)
    drop = (
        L2**2
        * wavelength**2
        * sc.norm(gravity)
        * (sc.constants.m_n**2 / (2 * sc.constants.h**2))
    )
    dropped_y = y + drop.to(unit=y.unit)

    two_theta = sc.atan2(y=sc.sqrt(x**2 + dropped_y**2), x=z)
    phi = sc.atan2(y=dropped_y, x=x)
    return {'two_theta': two_theta, 'phi': phi}


def test_scattering_angles_with_gravity_beams_unaligned_with_lab_coords():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    # Gravity and incident_beam are not aligned with the coordinate system
    # but orthogonal to each other.
    gravity = sc.vector([-0.3, -9.81, 0.01167883211678832], unit='m/s^2')
    incident_beam = sc.vector([1.6, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[1.8, 2.5, 3.6], [-0.4, -1.7, 2.9]], unit='m'
    )

    original_wavelength = wavelength.copy()
    original_gravity = gravity.copy()
    original_incident_beam = incident_beam.copy()
    original_scattered_beam = scattered_beam.copy()

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = _reference_scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'],
        expected['two_theta'].transpose(res['two_theta'].dims),
    )
    sc.testing.assert_allclose(
        res['phi'],
        expected['phi'].transpose(res['phi'].dims),
    )

    sc.testing.assert_identical(wavelength, original_wavelength)
    sc.testing.assert_identical(gravity, original_gravity)
    sc.testing.assert_identical(incident_beam, original_incident_beam)
    sc.testing.assert_identical(scattered_beam, original_scattered_beam)


def test_scattering_angles_with_gravity_binned_data():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    wavelength = sc.bins(
        dim='wavelength',
        data=wavelength,
        begin=sc.array(dims=['det'], values=[0, 2], unit=None),
        end=sc.array(dims=['det'], values=[2, 3], unit=None),
    )
    gravity = sc.vector([-0.3, -9.81, 0.01167883211678832], unit='m/s^2')
    incident_beam = sc.vector([1.6, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[1.8, 2.5, 3.6], [-0.4, -1.7, 2.9]], unit='m'
    )

    original_wavelength = wavelength.copy()
    original_gravity = gravity.copy()
    original_incident_beam = incident_beam.copy()
    original_scattered_beam = scattered_beam.copy()

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = _reference_scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(
        res['two_theta'],
        expected['two_theta'].transpose(res['two_theta'].dims),
    )
    sc.testing.assert_allclose(res['phi'], expected['phi'].transpose(res['phi'].dims))

    sc.testing.assert_identical(wavelength, original_wavelength)
    sc.testing.assert_identical(gravity, original_gravity)
    sc.testing.assert_identical(incident_beam, original_incident_beam)
    sc.testing.assert_identical(scattered_beam, original_scattered_beam)


@pytest.mark.parametrize('g', [[0.0, -9.81, 0.0], [2.3, 0.4, -7.5]])
def test_scattering_angles_with_gravity_uses_wavelength_dtype(g: list[float]):
    wavelength = sc.array(
        dims=['wavelength'], values=[1.6, 0.7], unit='Å', dtype='float32'
    )
    gravity = sc.vector(g, unit='m/s^2')
    incident_beam = sc.vector([0.0, 0.0, 41.1], unit='m')
    scattered_beam = sc.vector([1.8, 2.5, 3.6], unit='m')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    assert res['two_theta'].dtype == 'float32'
    assert res['phi'].dtype == 'float32'


@pytest.mark.parametrize('g', [[0.0, -9.81, 0.0], [2.3, 0.4, -7.5]])
def test_scattering_angles_with_gravity_supports_mismatching_units(g: list[float]):
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.7], unit='Å')
    gravity = sc.vector(g, unit='m/ms^2')
    incident_beam = sc.vector([0.0, 0.0, 410], unit='cm')
    scattered_beam = sc.vector([180, 1800, 2400], unit='mm')

    res = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    expected = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam.to(unit='m'),
        scattered_beam=scattered_beam.to(unit='m'),
        wavelength=wavelength,
        gravity=gravity.to(unit='m/s^2'),
    )

    sc.testing.assert_allclose(
        res['two_theta'], expected['two_theta'], rtol=sc.scalar(1e-6)
    )
    sc.testing.assert_allclose(res['phi'], expected['phi'])


@given(rotation=rotations())
@settings(max_examples=10)
def test_scattering_angles_with_gravity_invariant_under_rotation(
    rotation: sc.Variable,
) -> None:
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    # Gravity and incident_beam are not aligned with the coordinate system
    # but orthogonal to each other.
    gravity = sc.vector([-0.3, -9.81, 0.01167883211678832], unit='m/s^2')
    incident_beam = sc.vector([1.6, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[1.8, 2.5, 3.6], [-0.4, -1.7, 2.9]], unit='m'
    )
    original = beamline.scattering_angles_with_gravity(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    rotated = beamline.scattering_angles_with_gravity(
        incident_beam=rotation * incident_beam,
        scattered_beam=rotation * scattered_beam,
        wavelength=wavelength,
        gravity=rotation * gravity,
    )

    sc.testing.assert_allclose(original['two_theta'], rotated['two_theta'])
    sc.testing.assert_allclose(original['phi'], rotated['phi'])


def test_scattering_angle_in_yz_plane_requires_gravity_orthogonal_to_incident_beam():
    incident_beam = sc.vector([0.564, 1.2, -10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'], values=[[13, 24, 35], [51, -42, 33]], unit='m'
    )
    wavelength = sc.array(dims=['wavelength'], values=[1.2, 1.6, 1.8], unit='Å')
    gravity = sc.vector([0, 0, sc.constants.g.value], unit=sc.constants.g.unit)

    with pytest.raises(
        ValueError, match='`gravity` and `incident_beam` must be orthogonal'
    ):
        beamline.scattering_angle_in_yz_plane(
            incident_beam=incident_beam,
            scattered_beam=scattered_beam,
            wavelength=wavelength,
            gravity=gravity,
        )


def test_scattering_angle_in_yz_plane_small_gravity():
    # This case is unphysical but tests the consistency with `two_theta`.
    # Note that the scattered beam must be in the x-z plane for `two_theta`
    # and `scattering_angle_from_sample` to compute the same angle.
    incident_beam = sc.vector([0.0, 0.0, 10.4], unit='m')
    scattered_beam = sc.vectors(
        dims=['beam'],
        values=[[0, 24, 0], [0, -42, 0]],
        unit='m',
    )
    wavelength = sc.array(dims=['wavelength'], values=[1.2, 1.6, 1.8], unit='Å')
    gravity = sc.vector([0, -1e-11, 0], unit=sc.constants.g.unit)

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    ).broadcast(dims=['wavelength', 'beam'], shape=[3, 2])
    sc.testing.assert_allclose(res, expected)


@pytest.mark.parametrize('polar', [np.pi / 3, np.pi / 2, 2 * np.pi / 3, np.pi])
def test_scattering_angle_in_yz_plane_reproduces_polar_angle(
    polar: float,
):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.

    gravity = sc.vector([0.0, -1e-11, 0.0], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot1 = sc.spatial.rotations_from_rotvecs(sc.vector([-polar, 0, 0], unit='rad'))
    scattered_beam = rot1 * incident_beam

    wavelength = sc.scalar(1e-6, unit='Å')

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(res, sc.scalar(polar, unit='rad'))


@pytest.mark.parametrize('polar', [np.pi / 3, np.pi / 2, 2 * np.pi / 3, np.pi])
@pytest.mark.parametrize('x', [0.5, 11.4, -9.7])
def test_scattering_angle_in_yz_plane_does_not_depend_on_x(polar: float, x: float):
    # This case is unphysical but tests that the function reproduces
    # the expected angles using a rotated vector.

    gravity = sc.vector([0.0, -1e-11, 0.0], unit='cm/s^2')
    incident_beam = sc.vector([0.0, 0.0, 968.0], unit='cm')

    # With this definition, the x-axis has azimuthal=0.
    rot1 = sc.spatial.rotations_from_rotvecs(sc.vector([-polar, 0, 0], unit='rad'))
    scattered_beam_ref = rot1 * incident_beam
    scattered_beam_shift = scattered_beam_ref + sc.vector([x, 0.0, 0.0], unit='cm')

    wavelength = sc.scalar(1e-6, unit='Å')

    res_shift = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam_shift,
        wavelength=wavelength,
        gravity=gravity,
    )
    res_ref = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam_ref,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(res_shift, res_ref)


def test_scattering_angle_in_yz_plane_drops_in_expected_direction():
    wavelength = sc.scalar(1.6, unit='Å')
    gravity = sc.vector([0.0, -sc.constants.g.value, 0.0], unit=sc.constants.g.unit)
    incident_beam = sc.vector([0.0, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[0.0, 2.5, 8.6], [0.0, -1.7, 6.9]], unit='m'
    )

    with_gravity = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    without_gravity = beamline.two_theta(
        incident_beam=incident_beam, scattered_beam=scattered_beam
    )

    # The neutron was detected above the incident beam.
    # So using straight paths, it looks like it scattered at a
    # smaller angle (detected at a lower y) than in the real case with gravity.
    assert sc.all(with_gravity[0] > without_gravity[0]).value
    # The neutron was detected below the incident beam.
    # So the opposite of the above comment applies.
    assert sc.all(with_gravity[1] < without_gravity[1]).value


def test_scattering_angle_in_yz_plane_beams_aligned_with_lab_coords():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    gravity = sc.vector([0.0, -sc.constants.g.value, 0.0], unit=sc.constants.g.unit)
    incident_beam = sc.vector([0.0, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[0.0, 2.5, 3.6], [0.0, -1.7, 2.9]], unit='m'
    )

    original_wavelength = wavelength.copy()
    original_gravity = gravity.copy()
    original_incident_beam = incident_beam.copy()
    original_scattered_beam = scattered_beam.copy()

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    L2 = sc.norm(scattered_beam)
    x = scattered_beam.fields.x
    y = scattered_beam.fields.y
    z = scattered_beam.fields.z
    drop = (
        L2**2
        * sc.norm(gravity)
        * wavelength**2
        * sc.constants.m_n**2
        / (2 * sc.constants.h**2)
    )
    drop = drop.to(unit=y.unit)
    true_y = y + drop
    true_scattered_beam = sc.spatial.as_vectors(x, true_y, z)
    expected_theta = sc.asin(abs(true_y) / sc.norm(true_scattered_beam))
    expected_theta = expected_theta.transpose(res.dims)

    sc.testing.assert_allclose(res, expected_theta)

    sc.testing.assert_identical(wavelength, original_wavelength)
    sc.testing.assert_identical(gravity, original_gravity)
    sc.testing.assert_identical(incident_beam, original_incident_beam)
    sc.testing.assert_identical(scattered_beam, original_scattered_beam)


def _reference_scattering_angle_in_yz_plane(
    incident_beam: sc.Variable,
    scattered_beam: sc.Variable,
    gravity: sc.Variable,
    wavelength: sc.Variable,
) -> sc.Variable:
    # This is a simplified, independently checked implementation.
    e_z = incident_beam / sc.norm(incident_beam)
    e_y = -gravity / sc.norm(gravity)

    y = sc.dot(scattered_beam, e_y)
    z = sc.dot(scattered_beam, e_z)

    L2 = sc.norm(scattered_beam)
    drop = (
        L2**2
        * wavelength**2
        * sc.norm(gravity)
        * (sc.constants.m_n**2 / (2 * sc.constants.h**2))
    )
    dropped_y = y + drop.to(unit=y.unit)

    return sc.atan2(y=abs(dropped_y), x=z)


def test_scattering_angle_in_yz_plane_beams_unaligned_with_lab_coords():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    # Gravity and incident_beam are not aligned with the coordinate system
    # but orthogonal to each other.
    gravity = sc.vector([-0.3, -9.81, 0.01167883211678832], unit='m/s^2')
    incident_beam = sc.vector([1.6, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[1.8, 2.5, 3.6], [-0.4, -1.7, 2.9]], unit='m'
    )

    original_wavelength = wavelength.copy()
    original_gravity = gravity.copy()
    original_incident_beam = incident_beam.copy()
    original_scattered_beam = scattered_beam.copy()

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = _reference_scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(res, expected.transpose(res.dims))

    sc.testing.assert_identical(wavelength, original_wavelength)
    sc.testing.assert_identical(gravity, original_gravity)
    sc.testing.assert_identical(incident_beam, original_incident_beam)
    sc.testing.assert_identical(scattered_beam, original_scattered_beam)


def test_scattering_angle_in_yz_plane_binned_data():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.9, 0.7], unit='Å')
    wavelength = sc.bins(
        dim='wavelength',
        data=wavelength,
        begin=sc.array(dims=['det'], values=[0, 2], unit=None),
        end=sc.array(dims=['det'], values=[2, 3], unit=None),
    )
    gravity = sc.vector([-0.3, -9.81, 0.01167883211678832], unit='m/s^2')
    incident_beam = sc.vector([1.6, 0.0, 41.1], unit='m')
    scattered_beam = sc.vectors(
        dims=['det'], values=[[1.8, 2.5, 3.6], [-0.4, -1.7, 2.9]], unit='m'
    )

    original_wavelength = wavelength.copy()
    original_gravity = gravity.copy()
    original_incident_beam = incident_beam.copy()
    original_scattered_beam = scattered_beam.copy()

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    expected = _reference_scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    sc.testing.assert_allclose(res, expected.transpose(res.dims))
    sc.testing.assert_identical(wavelength, original_wavelength)
    sc.testing.assert_identical(gravity, original_gravity)
    sc.testing.assert_identical(incident_beam, original_incident_beam)
    sc.testing.assert_identical(scattered_beam, original_scattered_beam)


def test_scattering_angle_in_yz_plane_uses_wavelength_dtype():
    wavelength = sc.array(
        dims=['wavelength'], values=[1.6, 0.7], unit='Å', dtype='float32'
    )
    gravity = sc.vector([0.0, -9.81, 0.0], unit='m/s^2')
    incident_beam = sc.vector([0.0, 0.0, 41.1], unit='m')
    scattered_beam = sc.vector([1.8, 2.5, 3.6], unit='m')

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )
    assert res.dtype == 'float32'


def test_scattering_angle_in_yz_plane_supports_mismatching_units():
    wavelength = sc.array(dims=['wavelength'], values=[1.6, 0.7], unit='Å')
    gravity = sc.vector([0.0, -9810, 0.0], unit='m/ms^2')
    incident_beam = sc.vector([0.0, 0.0, 410], unit='cm')
    scattered_beam = sc.vector([180, 1800, 2400], unit='mm')

    res = beamline.scattering_angle_in_yz_plane(
        incident_beam=incident_beam,
        scattered_beam=scattered_beam,
        wavelength=wavelength,
        gravity=gravity,
    )

    expected = _reference_scattering_angle_in_yz_plane(
        incident_beam=incident_beam.to(unit='m'),
        scattered_beam=scattered_beam.to(unit='m'),
        wavelength=wavelength,
        gravity=gravity.to(unit='m/s^2'),
    )

    sc.testing.assert_allclose(res, expected)


def test_beam_aligned_unit_vectors_axis_aligned_inputs():
    res = beamline.beam_aligned_unit_vectors(
        incident_beam=sc.vector([0.0, 0.0, 2.1], unit='mm'),
        gravity=sc.vector([0.0, -4.6, 0.0], unit='m/s/s'),
    )
    assert len(res) == 3
    sc.testing.assert_identical(res['beam_aligned_unit_x'], sc.vector([1.0, 0.0, 0.0]))
    sc.testing.assert_identical(res['beam_aligned_unit_y'], sc.vector([0.0, 1.0, 0.0]))
    sc.testing.assert_identical(res['beam_aligned_unit_z'], sc.vector([0.0, 0.0, 1.0]))


def test_beam_aligned_unit_vectors_complicated_inputs():
    incident_beam = sc.vector([3.1, -0.2, 23.6], unit='m')
    gravity = sc.vector([-0.01, -9.5, 3.2], unit='m/s/s')
    # Subtract projection of gravity onto incident_beam to make the vectors orthogonal.
    gravity -= sc.to_unit(
        sc.dot(incident_beam, gravity) * incident_beam / sc.norm(incident_beam) ** 2,
        'm/s/s',
    )
    res = beamline.beam_aligned_unit_vectors(
        incident_beam=incident_beam, gravity=gravity
    )
    assert len(res) == 3
    ex = res['beam_aligned_unit_x']
    ey = res['beam_aligned_unit_y']
    ez = res['beam_aligned_unit_z']
    sc.testing.assert_allclose(sc.norm(ex), sc.scalar(1.0))
    sc.testing.assert_allclose(sc.norm(ey), sc.scalar(1.0))
    sc.testing.assert_allclose(sc.norm(ez), sc.scalar(1.0))
    sc.testing.assert_allclose(sc.dot(ex, ey), sc.scalar(0.0), atol=sc.scalar(1e-16))
    sc.testing.assert_allclose(sc.dot(ey, ez), sc.scalar(0.0), atol=sc.scalar(1e-16))
    sc.testing.assert_allclose(sc.dot(ez, ex), sc.scalar(0.0), atol=sc.scalar(1e-16))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""Models for peaks and background."""

from __future__ import annotations

import abc
import math
from collections.abc import Iterable
from copy import deepcopy

import numpy as np
import scipp as sc


class Model(abc.ABC):
    """Abstract base class for fitting models.

    This class defines the basic interface for models by way of public methods.
    Subclasses should override the protected methods ``_call``, ``_guess``, and
    optionally ``_param_bounds`` instead of their public counterparts.
    """

    def __init__(self, *, param_names: Iterable[str], prefix: str = '') -> None:
        """Initialize a base model.

        Parameters
        ----------
        param_names:
            Names of parameters in arbitrary order.
            Does not include the prefix.
        prefix:
            Prefix used for model parameters in all user-facing data.
        """
        self._prefix = prefix
        self._param_names = set(param_names)
        self._prefixed_param_names = {prefix + name for name in self._param_names}

    @abc.abstractmethod
    def _call(self, x: sc.Variable, params: dict[str, sc.Variable]) -> sc.Variable:
        """Evaluate the model at a given independent variable and parameters.

        Parameters
        ----------
        x:
            Independent variable.
        params:
            Dict from parameter names to values.
            Names are given *without* prefix.

        Returns
        -------
        :
            Model evaluated at the given independent variable.
        """
        ...

    @abc.abstractmethod
    def _guess(self, x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
        """Roughly estimate the model parameters.

        Parameters
        ----------
        x:
            Independent variable.
        y:
            Dependent variable.

        Returns
        -------
        :
            Estimated parameters.
            Dict keys are parameter names *without* the prefix.
        """
        ...

    def _param_bounds(self) -> dict[str, tuple[float, float]]:
        """Return bounds for parameters.

        Returns
        -------
        :
            Upper and lower bounds for parameters.
            Dict keys are parameter names *without* the prefix.
            Parameters omitted from this dict are unbounded.
        """
        return {}

    @property
    def prefix(self) -> str:
        """Prefix for parameter names."""
        return self._prefix

    @property
    def param_names(self) -> set[str]:
        """Parameter names including the prefix."""
        return self._prefixed_param_names.copy()

    def __call__(self, x: sc.Variable, **params: sc.Variable) -> sc.Variable:
        """Evaluate the model.

        Parameters
        ----------
        x:
            Independent variable.
        params:
            Parameter values.

        Returns
        -------
        :
            Model evaluated at the given independent variable and parameters.
        """
        if params.keys() != self.param_names:
            raise ValueError(
                f'Bad parameters for model {self.__class__.__name__}, '
                f'got: {set(params.keys())}, expected {self.param_names}'
            )
        return self._call(
            x, {name[len(self._prefix) :]: val for name, val in params.items()}
        )

    def guess(
        self, data: sc.DataArray, *, coord: str | None = None
    ) -> dict[str, sc.Variable]:
        """Roughly estimate the model parameters for given data.

        The estimate can be used as the starting point for a fit
        but does not necessarily represent a good fit by itself.

        Parameters
        ----------
        data:
            Data array where ``data.data`` is the dependent variable
            and a chosen coord (see below) is the independent variable.
        coord:
            Coordinate name of ``data`` to use as independent variable.
            If not given, ``data.dim`` is used instead.

        Returns
        -------
        :
            Estimated parameters.
        """
        if coord is None:
            coord = data.dim
        return {
            self._prefix + name: param
            for name, param in self._guess(x=data.coords[coord], y=data.data).items()
        }

    @property
    def param_bounds(self) -> dict[str, tuple[float, float]]:
        """Parameter bounds.

        Returns
        -------
        :
            Upper and lower bounds for parameters.
            Parameters omitted from this dict are unbounded.
        """
        return {
            self._prefix + name: bounds for name, bounds in self._param_bounds().items()
        }

    def fwhm(self, params: dict[str, sc.Variable]) -> sc.Variable:
        """Compute full width at half maximum.

        Note that this function is only implemented for peaked models!

        Parameters
        ----------
        params:
            Parameter values for which to compute the FWHM.

        Returns
        -------
        :
            FWHM of the model at the given parameters.

        Raises
        ------
        NotImplementedError:
            If this model does not support computing the FWHM.
        """
        raise NotImplementedError(
            f'FWHM is not implemented for model {self.__class__.__name__}'
        )

    def __add__(self, other: Model) -> CompositeModel:
        """Combine two models into a :class:`CompositeModel`."""
        if not isinstance(other, Model):
            return NotImplemented
        return CompositeModel(left=self, right=other, prefix='')

    def with_prefix(self, prefix: str) -> Model:
        """Return a copy of the model with a new prefix."""
        model = deepcopy(self)
        model._prefix = prefix
        model._prefixed_param_names = {prefix + name for name in model._param_names}
        return model


class CompositeModel(Model):
    """A combination of two models.

    Composite models contain a "left" and a "right" submodel which are combined into

    .. math::

        f(x) = \\text{left}(x) + \\text{right}(x)

    Composite models can be constructed by adding models, e.g.,

    .. code-block:: python

        left = PolynomialModel(degree=2)
        right = GaussianModel()
        composite = left + right

    The parameters of the composite are the union of the component parameters.
    If there is a clash between the names of component models, they must
    be disambiguated by using prefixes.
    """

    def __init__(self, left: Model, right: Model, *, prefix: str = '') -> None:
        """Initialize a composite model.

        Parameters
        ----------
        left:
            Left component model.
        right:
            Right component model.
        prefix:
            Prefix for *all* model parameter names.
            It is prepended to the prefixes of the component models.
        """
        if left.param_names & right.param_names:
            raise ValueError(
                f'Model {left.__class__.__name__} and model {right.__class__.__name__} '
                'have overlapping parameter names: '
                f'{left.param_names & right.param_names}. '
                'Use prefixes to disambiguate.'
            )
        self._left = left
        self._right = right
        super().__init__(
            prefix=prefix, param_names=left.param_names | right.param_names
        )

    def _call(self, x: sc.Variable, params: dict[str, sc.Variable]) -> sc.Variable:
        left = self._left(x, **{name: params[name] for name in self._left.param_names})
        right = self._right(
            x, **{name: params[name] for name in self._right.param_names}
        )
        return left + right

    def _guess(self, x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
        data = sc.DataArray(y, coords={y.dim: x})
        return {
            **self._left.guess(data),
            **self._right.guess(data),
        }

    def _param_bounds(self) -> dict[str, tuple[float, float]]:
        return self._left.param_bounds | self._right.param_bounds


class PolynomialModel(Model):
    """A polynomial of fixed degree.

    ``PolynomialModel(degree=n)`` implements

    .. math::

        f(x; a_0, \\ldots, a_n) = \\sum_{i=0}^{n}\\,a_i x^i

    where the sum is inclusive on the upper bound.
    :math:`a_i` are the parameters and are named ``['a0', 'a1', ...]``.
    """

    def __init__(self, *, degree: int, prefix: str = '') -> None:
        """Initialize a polynomial model.

        Parameters
        ----------
        degree:
            Degree of the polynomial.
        prefix:
            Prefix for model parameter names.
        """
        if degree <= 0:
            raise ValueError(f'Degree must be positive, got: {degree}')
        super().__init__(
            prefix=prefix, param_names=(f'a{i}' for i in range(degree + 1))
        )

    @property
    def degree(self) -> int:
        """The degree of the polynomial."""
        return len(self._param_names) - 1

    def _call(self, x: sc.Variable, params: dict[str, sc.Variable]) -> sc.Variable:
        a_degree = params[f'a{self.degree}']
        val = sc.full(value=a_degree.value, unit=a_degree.unit, sizes=x.sizes)
        for i in range(self.degree - 1, -1, -1):
            val *= x
            val += params[f'a{i}']
        return val

    def _guess(self, x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
        poly = np.polynomial.Polynomial.fit(x.values, y.values, deg=self.degree)
        return {
            f'a{i}': sc.scalar(c, unit=y.unit / x.unit**i)
            for i, c in enumerate(poly.convert().coef)
        }


class GaussianModel(Model):
    r"""A Gaussian function with arbitrary normalization.

    The model implements

    .. math::

        f(x; A, \mu, \sigma) = \frac{A}{\sqrt{2\pi}\sigma}
          \exp{\left(-\frac{{(x-\mu)}^2}{2\sigma^2}\right)}

    with parameters

    - :math:`A`: ``'amplitude'``
    - :math:`\mu`: ``'loc'``
    - :math:`\sigma`: ``'scale'``
    """

    def __init__(self, *, prefix: str = '') -> None:
        """Initialize a Gaussian model.

        Parameters
        ----------
        prefix:
            Prefix for model parameter names.
        """
        super().__init__(prefix=prefix, param_names=('amplitude', 'loc', 'scale'))

    def _call(self, x: sc.Variable, params: dict[str, sc.Variable]) -> sc.Variable:
        return _gaussian(x, **params)

    def _guess(self, x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
        params = _guess_from_peak(x, y)
        # Adjust by the normalization factor of gaussians.
        params['amplitude'] *= math.sqrt(2 * math.pi)
        return params

    def _param_bounds(self) -> dict[str, tuple[float, float]]:
        return {'scale': (0.0, np.inf)}

    def fwhm(self, params: dict[str, sc.Variable]) -> sc.Variable:
        """Compute full width at half maximum.

        Parameters
        ----------
        params:
            Parameter values for which to compute the FWHM.

        Returns
        -------
        :
            FWHM of the model at the given parameters.
        """
        return 2 * math.sqrt(2 * math.log(2)) * params[self._prefix + 'scale']


class LorentzianModel(Model):
    r"""A Lorentzian function with arbitrary normalization.

    The model implements

    .. math::

        f(x; A, \mu, \sigma) = \frac{A}{\pi} \frac{\sigma}{{(x-\mu)}^2 + \sigma^2}

    with parameters

    - :math:`A`: ``'amplitude'``
    - :math:`\mu`: ``'loc'``
    - :math:`\sigma`: ``'scale'``
    """

    def __init__(self, *, prefix: str = '') -> None:
        """Initialize a Lorentzian model.

        Parameters
        ----------
        prefix:
            Prefix for model parameter names.
        """
        super().__init__(prefix=prefix, param_names=('amplitude', 'loc', 'scale'))

    def _call(self, x: sc.Variable, params: dict[str, sc.Variable]) -> sc.Variable:
        return _lorentzian(x, **params)

    def _guess(self, x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
        params = _guess_from_peak(x, y)
        # Fudge factor taken from lmfit.
        # Not sure where exactly it comes from, but it is related to the normalization
        # of a Lorentzian and is approximately
        # 3.0 * math.pi / math.sqrt(2 * math.pi)
        params['amplitude'] *= 3.75
        return params

    def _param_bounds(self) -> dict[str, tuple[float, float]]:
        return {'scale': (0.0, np.inf)}

    def fwhm(self, params: dict[str, sc.Variable]) -> sc.Variable:
        """Compute full width at half maximum.

        Parameters
        ----------
        params:
            Parameter values for which to compute the FWHM.

        Returns
        -------
        :
            FWHM of the model at the given parameters.
        """
        return 2 * params[self._prefix + 'scale']


class PseudoVoigtModel(Model):
    r"""A Pseudo-Voigt function.

    The model implements

    .. math::

        f(x; A, \mu, \sigma, \alpha) = \alpha L(x; A, \mu, \sigma)
          + (1-\alpha) G(x; A, \mu, \sigma_G)

    where :math:`L` is a :class:`Lorentzian <LorentzianModel>`
    and :math:`G` is a :class:`Gaussian <GaussianModel>`.
    :math:`\sigma_G` is derived from :math:`\sigma` such that :math:`L` and
    :math:`G` have the same FWHM.

    It has parameters

    - :math:`A`: ``'amplitude'``
    - :math:`\mu`: ``'loc'``
    - :math:`\sigma`: ``'scale'``
    - :math:`\alpha`: ``'fraction'``
    """

    def __init__(self, *, prefix: str = '') -> None:
        super().__init__(
            prefix=prefix, param_names=('amplitude', 'loc', 'scale', 'fraction')
        )

    def _call(self, x: sc.Variable, params: dict[str, sc.Variable]) -> sc.Variable:
        params = dict(params.items())
        fraction = params.pop('fraction')

        lorentzian = _lorentzian(x, **params)
        # Adjust Gaussian scale such that Gaussian and Lorentzian have the same
        # FWHM of 2*scale.
        scale_g = params['scale'] / math.sqrt(2 * math.log(2))
        gaussian = _gaussian(
            x, amplitude=params['amplitude'], loc=params['loc'], scale=scale_g
        )

        return fraction * lorentzian + (1 - fraction) * gaussian

    def _guess(self, x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
        params = _guess_from_peak(x, y)
        params['fraction'] = sc.scalar(0.5)
        # See Lorentzian
        params['amplitude'] *= 3.75
        return params

    def _param_bounds(self) -> dict[str, tuple[float, float]]:
        return {'scale': (0.0, np.inf), 'fraction': (0.0, 1.0)}

    def fwhm(self, params: dict[str, sc.Variable]) -> sc.Variable:
        """Compute full width at half maximum.

        Parameters
        ----------
        params:
            Parameter values for which to compute the FWHM.

        Returns
        -------
        :
            FWHM of the model at the given parameters.
        """
        # Note that the Gaussian component has an adjusted scale to enable this.
        return 2 * params[self._prefix + 'scale']


def _gaussian(
    x: sc.Variable, *, amplitude: sc.Variable, loc: sc.Variable, scale: sc.Variable
) -> sc.Variable:
    # Avoid division by 0
    scale = sc.scalar(max(scale.value, 1e-15), variance=scale.variance, unit=scale.unit)

    val = x - loc
    val *= val
    val /= -2 * scale**2
    val = sc.exp(val, out=val)
    val *= amplitude / (math.sqrt(2 * math.pi) * scale)
    return val


def _lorentzian(
    x: sc.Variable, *, amplitude: sc.Variable, loc: sc.Variable, scale: sc.Variable
) -> sc.Variable:
    # Avoid division by 0
    scale = sc.scalar(max(scale.value, 1e-15), variance=scale.variance, unit=scale.unit)

    val = x - loc
    val *= val
    val += scale**2
    val = sc.reciprocal(val, out=val)
    val *= amplitude * scale / math.pi
    return val


def _guess_from_peak(x: sc.Variable, y: sc.Variable) -> dict[str, sc.Variable]:
    """Estimate the parameters of a peaked function.

    The estimation is based on a Gaussian but
    is good enough for similar functions, too.

    The function was adapted from ``lmfit.models.guess_from_peak``, see
    https://github.com/lmfit/lmfit-py/blob/e57aab2fe2059efc07535a67e4fdc577291e9067/lmfit/models.py#L42
    """
    y_min, y_max = sc.min(y), sc.max(y)

    # These are the points within FWHM of the peak.
    x_half_max = x[y > (y_max + y_min) / 2.0]
    if len(x_half_max) > 2:
        loc = x_half_max.mean()
        # Rough estimate of sigma ~ FWHM / 2.355 for Gaussian.
        # Exact gamma = FWHM / 2 for Lorentzian.
        scale = (x_half_max.max() - x_half_max.min()) / 2.0
    else:
        loc = x[np.argmax(y.values)]
        # 6.0 taken from lmfit, don't know where it comes from.
        scale = (sc.max(x) - sc.min(x)) / 6.0

    amplitude = scale * (y_max - y_min)
    return {
        'amplitude': amplitude,
        'loc': loc,
        'scale': scale,
    }
